import typing as t
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum, auto
from functools import partial
from typing import Any, Iterable, Union

from beet import Context, Function, Generator
from bolt import Runtime
from bolt.contrib.defer import Defer
from bolt.utils import internal
from mecha import AstChildren, AstCommand, AstRoot, Mecha
from mecha.contrib.nested_location import NestedLocationResolver
from pydantic import BaseModel

from .ast_converter import AstConverter
from .casting import TypeCaster
from .check import TypeChecker
from .optimizer import (
    ConstScoreManager,
    IrBranch,
    IrChildren,
    IrData,
    IrLiteral,
    IrOperation,
    IrScore,
    IrSet,
    IrSource,
    NbtValue,
    Optimizer,
    SourceTuple,
    TempDataManager,
    TempScoreManager,
    add_subtract_by_zero_removal,
    boolean_condition_propagation,
    branch_condition_propagation,
    commutative_set_collapsing,
    composite_literal_expansion,
    compound_match_data_compare,
    convert_cast,
    convert_data_arithmetic,
    convert_data_order_operation,
    convert_defined_boolean_condition,
    data_get_scaling,
    data_insert_score,
    data_set_scaling,
    data_string_propagation,
    deadcode_elimination,
    discard_casting,
    discard_non_numerical_casting,
    init_score_boolean_result,
    literal_to_constant_replacement,
    multiply_divide_by_fraction,
    multiply_divide_by_one_removal,
    noncommutative_set_collapsing,
    rename_temp_scores,
    set_and_get_cleanup,
    set_to_self_removal,
    source_copy_elision,
    store_result_inlining,
    store_set_data_compare,
)
from .typing import NbtTypeString
from .utils import identifier_generator, insert_nested_commands

__all__ = [
    "ExpressionOptions",
    "expression_options",
    "TempScoreManager",
    "ConstScoreManager",
    "ExpressionNode",
    "Expression",
]


class ExpressionOptions(BaseModel):
    """Bolt Expressions Options"""

    temp_objective: str = "bolt.expr.temp"
    const_objective: str = "bolt.expr.const"
    temp_storage: str = "bolt.expr:temp"
    init_path: str = "init_expressions"
    objective_prefix: str = ""
    const_score_prefix: str = "$"
    temp_score_prefix: str = "$"
    default_nbt_type: NbtTypeString = "int"
    default_floating_nbt_type: str = "double"

    disable_commands: bool = False


def expression_options(ctx: Context) -> ExpressionOptions:
    return ctx.validate("bolt_expressions", ExpressionOptions)


class ResultType(Enum):
    score = auto()
    data = auto()


@dataclass
class UnrollHelper:
    score_manager: TempScoreManager
    data_manager: TempDataManager

    ignored_sources: set[SourceTuple] = field(init=False, default_factory=set)
    temporaries: set[SourceTuple] = field(init=False, default_factory=set)
    data: dict[str, Any] = field(default_factory=dict)

    @contextmanager
    def provide(self, **kwargs: Any):
        prev_data = self.data
        self.data = {**self.data, **kwargs}

        yield self.data

        self.data = prev_data

    @contextmanager
    def ignore_source(self, source: SourceTuple):
        remove = source not in self.ignored_sources
        self.ignored_sources.add(source)

        yield

        if remove:
            self.ignored_sources.remove(source)

    def add_temporary(self, source: SourceTuple):
        self.temporaries.add(source)

    def remove_temporary(self, source: SourceTuple):
        self.temporaries.discard(source)

    def create_temporary(self, result: ResultType) -> IrSource:
        if result == ResultType.data:
            source = self.data_manager()
            self.add_temporary(source)
            return IrData(type=source.type, target=source.target, path=source.path)

        source = self.score_manager()
        self.add_temporary(source)
        return IrScore(holder=source.holder, obj=source.obj)


@dataclass(order=False, eq=False, kw_only=True)
class ExpressionNode(ABC):
    ctx: Union[Context, "Expression"] = field(repr=False)
    expr: "Expression" = field(init=False, repr=False)

    def __post_init__(self):
        if isinstance(self.ctx, Expression):
            self.expr = self.ctx
        else:
            self.expr = self.ctx.inject(Expression)

    @abstractmethod
    def unroll(
        self, helper: UnrollHelper
    ) -> tuple[Iterable[IrOperation], IrSource | IrLiteral]: ...


@dataclass(order=False, eq=False, kw_only=True)
class Unrolled(ExpressionNode):
    operations: Iterable[IrOperation] = ()
    value: IrSource | IrLiteral

    def unroll(self, helper: UnrollHelper):
        return self.operations, self.value


ResolveResult = SourceTuple | NbtValue | None


@dataclass(kw_only=True)
class LazyEntry:
    source: SourceTuple
    node: ExpressionNode
    commands: AstChildren[AstCommand]
    emit: bool = False


class Expression:
    ctx: Context
    opts: ExpressionOptions

    called_init: bool
    init_commands: list[str]
    commands: list[AstCommand] | None
    lazy_values: dict[SourceTuple, LazyEntry]

    type_caster: TypeCaster
    type_checker: TypeChecker
    optimizer: Optimizer
    ast_converter: AstConverter

    temp_score: TempScoreManager
    temp_data: TempDataManager
    const_score: ConstScoreManager
    identifiers: t.Generator[str, None, None]

    mecha: Mecha
    runtime: Runtime
    defer: Defer
    nested_location: NestedLocationResolver
    generator: Generator

    def __init__(self, ctx: Context):
        self.called_init = False
        self.init_commands = []
        self.commands = None
        self.lazy_values = {}

        self.ctx = ctx

        self.opts = self.ctx.inject(expression_options)
        self.identifiers = identifier_generator(ctx)
        self.mc = self.ctx.inject(Mecha)
        self.runtime = self.ctx.inject(Runtime)
        self.defer = self.ctx.inject(Defer)
        self.nested_location = self.ctx.inject(NestedLocationResolver)
        self.generator = self.ctx.generate

        self.temp_score = TempScoreManager(
            objective=self.opts.temp_objective,
            prefix=self.opts.temp_score_prefix,
            format=lambda _: self.opts.temp_score_prefix + next(self.identifiers),
        )
        self.temp_data = TempDataManager(
            "storage", self.opts.temp_storage, format=lambda _: next(self.identifiers)
        )
        self.const_score = ConstScoreManager(
            self.opts.const_objective, self.opts.const_score_prefix
        )

        self.type_caster = TypeCaster(ctx=self.ctx)
        self.type_checker = TypeChecker(ctx=self.ctx)

        self.optimizer = Optimizer(
            temp_score=self.temp_score,
            temp_data=self.temp_data,
            const_score=self.const_score,
            default_floating_nbt_type=self.opts.default_floating_nbt_type,
        )
        self.optimizer.add_rules(
            composite_literal_expansion=partial(
                composite_literal_expansion, opt=self.optimizer, ctx=self.ctx
            ),
            data_insert_score=data_insert_score,
            convert_cast=convert_cast,
            compound_match_data_compare=partial(
                compound_match_data_compare, opt=self.optimizer
            ),
            store_set_data_compare=partial(store_set_data_compare, opt=self.optimizer),
            convert_data_arithmetic=partial(convert_data_arithmetic, self.optimizer),
            convert_data_order_operation=partial(
                convert_data_order_operation, opt=self.optimizer
            ),
            discard_casting=discard_casting,
            # features
            data_set_scaling=partial(data_set_scaling, opt=self.optimizer),
            data_get_scaling=data_get_scaling,
            # cleanup
            multiply_divide_by_fraction=multiply_divide_by_fraction,
            multiply_divide_by_one_removal=multiply_divide_by_one_removal,
            add_subtract_by_zero_removal=add_subtract_by_zero_removal,
            noncommutative_set_collapsing=noncommutative_set_collapsing,
            commutative_set_collapsing=commutative_set_collapsing,
            data_string_propagation=data_string_propagation,
            literal_to_constant_replacement=partial(
                literal_to_constant_replacement, self.optimizer
            ),
            boolean_condition_propagation=boolean_condition_propagation,
            branch_condition_propagation=branch_condition_propagation,
            convert_defined_boolean_condition=partial(
                convert_defined_boolean_condition, opt=self.optimizer
            ),
            # typing
            type_caster=self.type_caster,
            type_checker=self.type_checker,
            # post type-checking cleanup
            discard_non_numerical_casting=discard_non_numerical_casting,
            source_copy_elision_post=partial(source_copy_elision, opt=self.optimizer),
            set_to_self_removal=set_to_self_removal,
            set_and_get_cleanup=set_and_get_cleanup,
            store_result_inlining=store_result_inlining,
            deadcode_elimination=partial(deadcode_elimination, opt=self.optimizer),
            init_score_boolean_result=init_score_boolean_result,
            rename_temp_scores=partial(rename_temp_scores, self.optimizer),
        )

        self.ast_converter = AstConverter(
            default_nbt_type=self.opts.default_nbt_type, mc=self.mc
        )

    def inject_command(self, *cmds: str | AstCommand):
        commands = self.commands
        if commands is None:
            commands = self.runtime.commands

        for cmd in cmds:
            if isinstance(cmd, str):
                cmd = self.mc.parse(cmd, using="command")

            commands.append(cmd)

    @contextmanager
    def scope(self, result: list[AstCommand] | None = None):
        if result is None:
            result = []

        prev = self.commands
        self.commands = result

        yield self.commands

        self.commands = prev

    @contextmanager
    def anonymous_function(
        self, template: str = "anonymous_{incr}"
    ) -> t.Generator[str, None, None]:
        namespace, resolved = self.nested_location.resolve()
        location = self.generator.format(f"{namespace}:{resolved}/{template}")

        with self.runtime.scope() as cmds:
            yield location

        cmd = self.mc.parse(f"execute function {location}:\n  ...", using="command")
        root = AstRoot(commands=AstChildren(cmds))
        self.runtime.commands.append(insert_nested_commands(cmd, root))

    def unroll(
        self, node: ExpressionNode
    ) -> tuple[Iterable[IrOperation], IrSource | IrLiteral, UnrollHelper]:
        helper = UnrollHelper(
            score_manager=self.temp_score, data_manager=self.temp_data
        )
        operations, result = node.unroll(helper)

        if isinstance(result, IrSource):
            helper.remove_temporary(result.to_tuple())

        return operations, result, helper

    @internal
    def resolve(self, node: ExpressionNode, lazy: bool = False) -> SourceTuple:
        operations, result, helper = self.unroll(node)

        if not isinstance(result, IrSource):
            score = self.optimizer.generate_score()
            operations = (IrSet(left=score, right=result),)
            result = score

        source = result.to_tuple()

        if source in self.lazy_values:
            del self.lazy_values[source]

        nodes, _ = self.optimizer(operations, temporaries=helper.temporaries)
        cmds = self.ast_converter(nodes)

        if not lazy:
            self.inject_command(*cmds)
            return source

        entry = LazyEntry(source=source, node=node, commands=cmds)
        self.lazy_values[source] = entry
        self.defer(partial(self.emit_lazy, entry=entry))

        return source

    @contextmanager
    @internal
    def resolve_branch(self, node: ExpressionNode):
        operations, result, helper = self.unroll(node)

        if not isinstance(result, IrSource):
            return

        result_tuple = result.to_tuple()

        if result_tuple in self.lazy_values:
            del self.lazy_values[result_tuple]

        nodes, temporaries = self.optimizer(
            operations,
            temporaries=helper.temporaries,
            rename_temp_scores=False,
            deadcode_elimination=False,
        )

        with self.runtime.scope() as cmds:
            yield

        branch = IrBranch(target=result, children=IrChildren.from_ast(cmds))

        nodes, _ = self.optimizer(
            (*nodes, branch),
            disable_all=True,
            temporaries=(*temporaries, result_tuple),
            branch_condition_propagation=True,
            convert_defined_boolean_condition=True,
            rename_temp_scores=True,
            deadcode_elimination=True,
        )

        cmds = self.ast_converter(nodes)
        self.inject_command(*cmds)

    def unroll_lazy(
        self, source: SourceTuple, helper: UnrollHelper
    ) -> tuple[Iterable[IrOperation], IrSource | IrLiteral] | None:
        if source in helper.ignored_sources:
            return None

        if helper.data.get("ignore_lazy"):
            return None

        if entry := self.lazy_values.get(source):
            helper.add_temporary(source)

            with helper.ignore_source(source):
                return entry.node.unroll(helper)

        return None

    def evaluate_lazy(self, source: SourceTuple):
        if entry := self.lazy_values.get(source):
            entry.emit = True
            del self.lazy_values[source]

    def emit_lazy(self, entry: LazyEntry):
        if entry.emit:
            self.inject_command(*entry.commands)

    def init(self):
        """Injects a function which creates `ConstantSource` fakeplayers"""
        path = self.ctx.generate.path(self.opts.init_path)
        self.inject_command(f"function {path}")
        self.called_init = True

    def generate_init(self) -> Function:
        function = Function(
            self.init_commands,
            prepend_tags=["minecraft:load"] if not self.called_init else None,
        )

        if not self.init_commands:
            return function

        self.ctx.generate(self.opts.init_path, function)

        return function



from types import GenericAlias, NoneType, UnionType
from typing import (
    Any,
    Iterable,
    Literal,
    TypedDict,
    TypeGuard,
    Union,
    _UnionGenericAlias,  # type: ignore
    cast,
    get_args,
    get_origin,
    get_type_hints,
    is_typeddict,
)

from beet import Context
from nbtlib import (
    Array,
    Byte,
    Compound,
    CompoundMatch,
    Double,
    End,
    Float,
    Int,
    List,
    ListIndex,
    Long,
    NamedKey,
    Short,
    String,
)  # type: ignore

from .utils import format_name, get_globals, type_name  # type: ignore

__all__ = ["is_type", "convert_type", "literal_types"]


NBT_TYPE_STRING = ("byte", "short", "int", "long", "float", "double")
NbtTypeString = Literal["byte", "short", "int", "long", "float", "double"]


NumericNbtValue = Byte | Short | Int | Long | Float | Double
NbtValue = NumericNbtValue | String | List | Array | Compound

NbtType = Union[
    # numeric
    type[NumericNbtValue],
    # string
    type[str],
    # array
    type[Array],
    # lists
    type[list["NbtType"]],
    # compounds
    # type[TypedDict],
    type[dict[str, "NbtType"]],
    dict[str, "NbtType"],
    # unions
    UnionType,
    # any
    type[Any],
]


literal_types: dict[str, type[NbtValue]] = {
    "byte": Byte,
    "short": Short,
    "int": Int,
    "long": Long,
    "float": Float,
    "double": Double,
    "compound": Compound,
    "list": List,
    "string": String,
}


def format_type(t: Any, *, __refs: list[Any] | None = None) -> str:
    if __refs is None:
        __refs = []

    circular_ref = t in __refs
    __refs.append(t)

    if t in (None, NoneType):
        return "None"

    if isinstance(t, (UnionType, _UnionGenericAlias)):
        return " | ".join(format_type(x, __refs=__refs) for x in get_args(t))

    if isinstance(t, GenericAlias):
        origin = format_type(get_origin(t), __refs=__refs)
        args = (format_type(x, __refs=__refs) for x in get_args(t))

        if circular_ref:
            return f"{origin}[...]"

        return f"{origin}[{', '.join(args)}]"

    if is_typeddict_guard(t) and t.__name__ == "__anonymous_dict__":
        t = get_dict_fields(t)
    if isinstance(t, dict):
        t_dict = cast(dict[str, Any], t)

        if circular_ref:
            return "{...}"

        return (
            "{"
            + ", ".join(
                f"{key}: {format_type(val, __refs=__refs)}"
                for key, val in t_dict.items()
            )
            + "}"
        )

    return format_name(t)


def convert_tag(value: Any) -> NbtValue | None:
    match value:
        case (
            Byte()
            | Short()
            | Int()
            | Float()
            | Double()
            | String()
            | List()
            | Array()
            | Compound()
        ):
            return value
        case list():
            value = cast(list[Any], value)
            elements = [convert_tag(v) for v in value]
            if any(el is None for el in elements):
                return None
            return List(elements)
        case dict(dict_value):  # type: ignore
            dict_value = cast(dict[str, Any], dict_value)
            items = [(key, convert_tag(value)) for key, value in dict_value.items()]
            if any(value is None for _, value in items):
                return None
            return Compound(items)  # type: ignore
        case bool():
            return Byte(value)
        case float():
            return Float(value)
        case int():
            return Int(value)
        case str():
            return String(value)
        case _:
            return None


def is_typeddict_guard(value: Any) -> TypeGuard[type[TypedDict]]:
    return is_typeddict(value)


def is_alias(value: Any, origin: type | tuple[type, ...] | None = None) -> bool:
    if not isinstance(value, GenericAlias):
        return False

    if origin is not None:
        value_origin = get_origin(value)

        if not isinstance(value_origin, type):
            return False

        return issubclass(value_origin, origin)

    return True


def is_union(value: Any) -> bool:
    return isinstance(value, (UnionType, _UnionGenericAlias))


def is_optional(value: Any) -> bool:
    if not is_union(value):
        return False

    return NoneType in get_args(value)


def is_numeric_type(value: Any) -> TypeGuard[type[NumericNbtValue]]:
    return isinstance(value, type) and issubclass(value, (int, float))


def is_string_type(value: Any) -> TypeGuard[type[String]]:
    return isinstance(value, type) and issubclass(value, str)


def is_list_type(value: Any) -> TypeGuard[type[list[Any]]]:
    if is_alias(value):
        return is_alias(value, list)

    if not isinstance(value, type):
        return False

    return issubclass(value, list)


def is_array_type(value: Any) -> TypeGuard[type[Array]]:
    return isinstance(value, type) and issubclass(value, Array)


def is_compound_alias(value: Any) -> TypeGuard[type[dict[str, Any]]]:
    if is_alias(value):
        return is_alias(value, dict)

    return isinstance(value, type) and issubclass(value, dict)


def is_fixed_compound(value: Any) -> TypeGuard[dict[str, Any] | type[TypedDict]]:
    return isinstance(value, dict) or is_typeddict(value)


def is_compound_type(value: Any) -> TypeGuard[type[dict[str, Any]] | type[TypedDict]]:
    return is_compound_alias(value) or is_fixed_compound(value)


def unwrap_optional_type(value: Any) -> Any:
    if not is_union(value):
        return value

    args = get_args(value)

    if NoneType not in args:
        return value

    return Union[tuple(v for v in args if v is not NoneType)]  # type: ignore


def get_dict_fields(
    t: type[TypedDict] | dict[str, Any], ctx: Context | None = None
) -> dict[str, NbtType]:
    globalns = None

    if isinstance(t, dict):
        fields = t
    else:
        globalns = get_globals(t, ctx)
        fields = get_type_hints(t, globalns=globalns)

    result: dict[str, NbtType] = {}

    for key, value in fields.items():
        value_type = convert_type(value, globalns=globalns)
        result[key] = value_type if value_type is not None else Any

    return dict(result)


def is_type(value: Any, allow_dict: bool = True) -> TypeGuard[NbtType]:
    if allow_dict and isinstance(value, dict):
        value = cast(dict[Any, Any], value)
        return any(is_type(v) for v in value.values())

    return value in (Any, None) or isinstance(
        value, (type, UnionType, _UnionGenericAlias, GenericAlias)
    )


def convert_type(
    value: Any, is_origin: bool = False, globalns: dict[str, Any] | None = None
) -> NbtType | None:
    if value is Any:
        return Any

    if value in (None, NoneType):
        return None

    if isinstance(value, dict):
        value = cast(dict[str, Any], value)

        type_dict: dict[str, NbtType] = {}
        for key, val in value.items():
            val_type = convert_type(val, globalns=globalns)
            type_dict[key] = val_type if val_type is not None else Any

        return TypedDict("__anonymous_dict__", type_dict)  # type: ignore

    if is_typeddict(value):
        return value  # type: ignore

    if is_alias(value, Compound):
        args = get_args(value)
        return convert_type(dict[str, args[0]], globalns=globalns)  # type: ignore

    if isinstance(value, (UnionType, _UnionGenericAlias)):
        args = get_args(value)
        converted = tuple(convert_type(arg, globalns=globalns) for arg in args)

        return Union[converted]  # type: ignore

    if isinstance(value, GenericAlias):
        args = get_args(value)

        converted = tuple(convert_type(arg, globalns=globalns) for arg in args)
        origin = convert_type(value.__origin__, is_origin=True)

        if isinstance(origin, type) and issubclass(origin, Compound):
            converted = (converted[-1],)

        if len(converted) == 1:
            converted = converted[0]

        return origin[converted]  # type: ignore

    if isinstance(value, type):
        if issubclass(value, (Byte, Short, Long, Double, Array)):
            return value
        if issubclass(value, bool):
            return Byte
        if issubclass(value, (float, Float)):
            return Float
        if issubclass(value, (int, Int)):
            return Int
        if issubclass(value, str):
            return String
        if issubclass(value, dict):
            return dict if is_origin else convert_type(dict[str, Any])  # type: ignore
        if issubclass(value, List):
            subtype = Any if value.subtype is End else value.subtype
            return list[convert_type(subtype)]  # type: ignore
        if issubclass(value, list):
            return list if is_origin else list[Any]  # type: ignore

    if isinstance(value, str) and globalns is not None and value in globalns:
        return globalns[value]

    raise TypeError(f"type {type_name(value)} cannot be converted to nbt type.")


Accessor = Union[NamedKey, ListIndex, CompoundMatch]


def access_type_by_path(
    t: NbtType | None, path: Iterable[Accessor], ctx: Context | None = None
) -> NbtType | None:
    new_accessors: tuple[Accessor, ...] = tuple(path)

    for accessor in new_accessors:
        t = access_type(t, accessor, ctx)

    return t


class A(TypedDict):
    a: str
    b: int


def infer_dict(value: dict[str, NbtValue]) -> NbtType | None:
    return convert_type({key: infer_type(val) for key, val in value.items()})


def infer_list(value: list[NbtValue]) -> NbtType:
    if not len(value):
        return list[Any]

    options = tuple(infer_type(element) for element in value)

    return list[Union[options]]  # type: ignore


def infer_type(value: Any, shallow: bool = False) -> NbtType | None:
    if isinstance(value, dict):
        if shallow:
            return dict[str, Any]

        value = cast(dict[str, Any], value)
        return infer_dict(value)

    if isinstance(value, list):
        if shallow:
            return list[Any]

        value = cast(list[Any], value)
        return infer_list(value)

    return convert_type(type(value))


def access_typeddict(
    t: type[TypedDict], accessor: Accessor, ctx: Context | None = None
) -> NbtType | None:
    if isinstance(accessor, CompoundMatch):
        return t

    if isinstance(accessor, NamedKey):
        key = accessor.key

        fields = get_dict_fields(t, ctx)

        if attr_type := fields.get(key):
            result = convert_type(attr_type)
            return result if result is not None else Any

    return None


def access_compound_alias(
    t: type[dict[str, Any]], accessor: Accessor
) -> NbtType | None:
    if isinstance(accessor, CompoundMatch):
        return t

    if isinstance(accessor, NamedKey):
        args = get_args(t)

        if not args:
            return Any

        return args[-1]

    return None


def access_list(t: type[list[Any]], accessor: Accessor) -> NbtType | None:
    if isinstance(accessor, ListIndex):
        if is_alias(t, list):
            arg = get_args(t)
            return arg[0] if arg else Any

        if isinstance(t, type) and issubclass(t, List):
            if t.subtype is End:
                return Any
            return t.subtype

    return None


def access_array(t: type[Array], accessor: Accessor) -> NbtType | None:
    if isinstance(accessor, ListIndex):
        wrapper = t.wrapper
        return wrapper if wrapper is not None else Any

    return None


def access_type(
    current_type: NbtType | None, accessor: Accessor, ctx: Context | None = None
) -> NbtType | None:
    if current_type in (Any, None):
        return current_type

    if is_numeric_type(current_type) or is_string_type(current_type):
        return None

    if is_union(current_type):
        args = get_args(current_type)
        subtypes = tuple(access_type(arg, accessor) for arg in args)

        return convert_type(Union[subtypes])  # type: ignore

    if is_typeddict_guard(current_type):
        return access_typeddict(current_type, accessor, ctx)

    if is_compound_alias(current_type):
        return access_compound_alias(current_type, accessor)

    if is_list_type(current_type):
        return access_list(current_type, accessor)

    if is_array_type(current_type):
        return access_array(current_type, accessor)

    return None



from abc import ABC
from bisect import bisect_left
from contextlib import contextmanager
from dataclasses import dataclass, field, replace
from enum import Enum
from fractions import Fraction
from types import TracebackType
from typing import (
    Any,
    Callable,
    Concatenate,
    Generator,
    Generic,
    Iterable,
    Iterator,
    Literal,
    NamedTuple,
    ParamSpec,
    TypeGuard,
    TypeVar,
    Union,
    cast,
)

from beet import Context
from bolt.utils import internal
from mecha import AbstractChildren, AbstractNode, AstNode
from nbtlib import (  # type:ignore
    Compound,
    CompoundMatch,
    Double,
    Float,
    Int,
    List,
    ListIndex,
    NamedKey,
    Numeric,
    Path,
    String,
)

from .typing import (
    Accessor,
    NbtType,
    NbtValue,
    access_type_by_path,
    convert_tag,
    infer_type,
    is_array_type,
    is_compound_type,
    is_list_type,
    is_numeric_type,
    is_string_type,
    literal_types,
    unwrap_optional_type,
)

__all__ = [
    "Rule",
    "SmartRule",
    "SmartGenerator",
    "Optimizer",
    "use_smart_generator",
    "smart_generator",
    "noncommutative_set_collapsing",
    "commutative_set_collapsing",
    "data_set_scaling",
    "data_get_scaling",
    "multiply_divide_by_fraction",
    "multiply_divide_by_one_removal",
    "add_subtract_by_zero_removal",
    "set_to_self_removal",
    "set_and_get_cleanup",
    "literal_to_constant_replacement",
    "DataTargetType",
    "IrNode",
    "IrSource",
    "IrScore",
    "IrLiteral",
    "IrOperation",
    "IrUnary",
    "IrBinary",
    "is_op",
    "is_unary",
    "is_binary",
]

T = TypeVar("T")


class IrNode(AbstractNode): ...


IrNodeType = TypeVar("IrNodeType", bound=IrNode, covariant=True)
AstNodeType = TypeVar("AstNodeType", bound=AstNode, covariant=True)


@dataclass(frozen=True, kw_only=True)
class IrRaw(IrNode, Generic[AstNodeType]):
    node: AstNodeType


class IrChildren(AbstractChildren[IrNodeType]):
    @classmethod
    def from_ast(
        cls, children: Iterable[AstNodeType]
    ) -> "IrChildren[IrRaw[AstNodeType]]":
        return IrChildren(IrRaw(node=node) for node in children)


class IrSource(IrNode, ABC):
    def to_tuple(self) -> "SourceTuple": ...


@dataclass(frozen=True, kw_only=True)
class IrScore(IrSource):
    holder: str
    obj: str

    def to_tuple(self) -> "ScoreTuple":
        return ScoreTuple(self.holder, self.obj)


class IrBoolScore(IrScore): ...


DataTargetType = Literal["storage", "entity", "block"]


@dataclass(frozen=True, kw_only=True)
class IrData(IrSource):
    type: DataTargetType
    target: str
    path: Path
    nbt_type: NbtType = Any
    scale: float | None = None

    def to_tuple(self) -> "DataTuple":
        return DataTuple(self.type, self.target, self.path)


@dataclass(frozen=True, kw_only=True)
class IrDataString(IrData):
    range: int | tuple[int | None, int | None]

    @property
    def normalized_range(self) -> tuple[int, int | None]:
        if isinstance(self.range, tuple):
            start, end = self.range
        else:
            start = self.range
            end = None if start == -1 else self.range + 1

        if start is None:
            start = 0

        return (start, end)

    def to_tuple(self) -> "StringDataTuple":  # type: ignore
        return StringDataTuple(self.type, self.target, self.path, self.normalized_range)


@dataclass(frozen=True, kw_only=True)
class IrLiteral(IrNode):
    value: NbtValue


CompositeNbtValue = Union[
    NbtValue,
    list[Union["CompositeNbtValue", IrSource]],
    dict[str, Union["CompositeNbtValue", IrSource]],
]


@dataclass(frozen=True, kw_only=True)
class IrCompositeLiteral(IrNode):
    value: CompositeNbtValue


@dataclass(frozen=True, kw_only=True)
class IrCondition(IrNode):
    op: str
    negated: bool = False


@dataclass(frozen=True, kw_only=True)
class IrUnaryCondition(IrCondition):
    target: IrSource


@dataclass(frozen=True, kw_only=True)
class IrBinaryCondition(IrCondition):
    left: IrSource | IrLiteral
    right: IrSource | IrLiteral


def is_condition(
    obj: Any, op: str | Iterable[str] | None = None
) -> TypeGuard[IrCondition]:
    if not isinstance(obj, IrCondition):
        return False

    if op is None:
        return True

    if isinstance(op, str):
        op = (op,)

    return obj.op in op


def is_unary_condition(
    obj: Any, op: str | Iterable[str] | None = None
) -> TypeGuard[IrUnaryCondition]:
    return is_condition(obj, op) and isinstance(obj, IrUnaryCondition)


def is_binary_condition(
    obj: Any, op: str | Iterable[str] | None = None
) -> TypeGuard[IrBinaryCondition]:
    return is_condition(obj, op) and isinstance(obj, IrBinaryCondition)


class StoreType(Enum):
    result = "result"
    success = "success"


@dataclass(frozen=True, kw_only=True)
class IrStore(IrNode):
    type: StoreType
    value: IrSource
    scale: float = 1
    cast_type: NbtType = Any


OperandType = IrLiteral | IrSource | IrCondition


@dataclass(frozen=True, kw_only=True)
class IrOperation(IrNode):
    op: str
    store: IrChildren[IrStore] = field(default_factory=IrChildren)

    destructive: bool = True

    @property
    def targets(self) -> tuple[IrSource, ...]:
        return tuple(s.value for s in self.store)

    @property
    def operands(self) -> tuple[IrSource | IrCondition | IrLiteral, ...]:
        return ()


@dataclass(frozen=True, kw_only=True)
class IrUnary(IrOperation):
    target: IrSource | IrCondition

    @property
    def targets(self):
        targets = super().targets

        if self.destructive and isinstance(self.target, IrSource):
            return targets + (self.target,)

        return targets

    @property
    def operands(self) -> tuple[IrSource | IrCondition]:
        return (self.target,)


@dataclass(frozen=True, kw_only=True)
class IrBinary(IrOperation):
    left: IrSource
    right: IrSource | IrCondition | IrLiteral

    uses_left: bool = True

    @property
    def targets(self):
        targets = super().targets

        if self.destructive:
            return targets + (self.left,)

        return targets

    @property
    def operands(self) -> tuple[IrSource | IrCondition | IrLiteral, ...]:
        if not self.uses_left:
            return (self.right,)

        return (self.left, self.right)


@dataclass(frozen=True, kw_only=True)
class IrGetLength(IrUnary):
    op: str = field(default="get_length", init=False)
    destructive: bool = field(default=False, init=False)


@dataclass(frozen=True, kw_only=True)
class IrInsert(IrBinary):
    op: str = field(default="insert", init=False)
    destructive: bool = field(default=True, init=False)
    index: int


@dataclass(frozen=True, kw_only=True)
class IrSet(IrBinary):
    op: str = field(default="set", init=False)
    destructive: bool = field(default=True, init=False)
    uses_left: bool = field(default=False, init=False)


@dataclass(frozen=True, kw_only=True)
class IrCast(IrBinary):
    op: str = field(default="cast", init=False)
    destructive: bool = field(default=True, init=False)
    uses_left: bool = field(default=False, init=False)
    cast_type: NbtType = Any
    scale: float = 1


@dataclass(frozen=True, kw_only=True)
class IrBranch(IrUnary):
    op: str = field(default="branch", init=False)
    destructive: bool = field(default=False, init=False)

    children: IrChildren[Any]


def is_op(obj: Any, op: str | Iterable[str] | None = None) -> TypeGuard[IrOperation]:
    if not isinstance(obj, IrOperation):
        return False

    if op is None:
        return True

    if isinstance(op, str):
        op = (op,)

    return obj.op in op


def is_unary(obj: Any, op: str | Iterable[str] | None = None) -> TypeGuard[IrUnary]:
    return is_op(obj, op) and isinstance(obj, IrUnary)


def is_binary(obj: Any, op: str | Iterable[str] | None = None) -> TypeGuard[IrBinary]:
    return is_op(obj, op) and isinstance(obj, IrBinary)


def is_copy_op(node: Any) -> TypeGuard[IrBinary]:
    if is_binary(node, "set"):
        return True

    if isinstance(node, IrCast):
        cast_type = unwrap_optional_type(node.cast_type)

        if node.scale != 1:
            return False

        if isinstance(node.left, IrScore) and isinstance(node.right, IrScore):
            return True

        if isinstance(node.left, IrData) and isinstance(node.right, IrData):
            if cast_type == unwrap_optional_type(node.right.nbt_type):
                return True
            if cast_type is Any:
                return True

        if isinstance(node.right, IrLiteral) and cast_type is Any:
            return True

        return False

    return False


SCORE_OPERATIONS = ("set", "add", "sub", "mul", "div", "mod", "min", "max")

DATA_OPERATIONS = ("set", "remove", "insert", "append", "prepend", "merge")


class SmartGenerator(Generator[T, None, None]):
    """Implements `.push(val)` which allows you to 'prepend' values to a generator.
    Allows you to peek values in the future, and return them to be consumed later.

    >>> gen = SmartGenerator(i for i in range(10))
    >>> val = next(gen)
    >>> gen.push(val)
    >>> next(gen)
    0
    """

    _values: Iterator[T]
    _pre: list[T]

    def __init__(self, values: Iterable[T]):
        self._values = iter(values)
        self._pre = []

    def __iter__(self):
        return self

    def __next__(self) -> T:
        if self._pre:
            return self._pre.pop()

        return next(self._values)

    def push(self, val: T | None):
        if val is not None:
            self._pre.append(val)

    def send(self, __value: None) -> T: ...

    def throw(
        self,
        __typ: BaseException | type[BaseException],
        __val: BaseException | object = None,
        __tb: TracebackType | None = None,
    ) -> T: ...


P = ParamSpec("P")
Rule = Callable[Concatenate[Iterable[T], P], Iterable[T]]

SmartRule = Callable[Concatenate[SmartGenerator[T], P], Iterable[T]]


def use_smart_generator(func: SmartRule[T, P]) -> Rule[T, P]:
    """
    Provides a `SmartGenerator` object as first argument to the decorated rule.
    Returned function still takes an `Iterable` object.
    """

    def rule(arg: Iterable[T], *args: P.args, **kwargs: P.kwargs) -> Iterable[T]:
        return func(SmartGenerator(arg), *args, **kwargs)

    return rule


def smart_generator(func: Rule[T, P]) -> Callable[[Iterable[T]], SmartGenerator[T]]:
    def rule(arg: Iterable[T], *args: P.args, **kwargs: P.kwargs):
        return SmartGenerator(func(arg, *args, **kwargs))

    return rule


class ScoreTuple(NamedTuple):
    holder: str
    obj: str


class DataTuple(NamedTuple):
    type: DataTargetType
    target: str
    path: Path


class StringDataTuple(NamedTuple):
    type: DataTargetType
    target: str
    path: Path
    range: tuple[int, int | None]


SourceTuple = ScoreTuple | DataTuple | StringDataTuple


@dataclass
class TempScoreManager:
    objective: str
    prefix: str

    counter: int
    format: Callable[[int], str]

    def __init__(
        self, objective: str, prefix: str, format: Callable[[int], str] | None = None
    ):
        self.objective = objective
        self.prefix = prefix
        self.format = format if format is not None else lambda n: f"{self.prefix}{n}"
        self.counter = 0

    def __call__(self) -> ScoreTuple:
        name = self.format(self.counter)
        self.counter += 1

        return ScoreTuple(name, self.objective)

    @contextmanager
    def override(self, format: Callable[[int], str] | None = None, reset: bool = False):
        counter = self.counter
        if reset:
            self.counter = 0

        prev_format = self.format
        if format:
            self.format = format

        yield

        self.counter = counter
        self.format = prev_format


@dataclass
class TempDataManager:
    target_type: DataTargetType
    target: str

    counter: int = field(default=0, init=False)
    format: Callable[[int], str] = field(default=lambda n: f"i{n}")  # type: ignore

    def __call__(self) -> DataTuple:
        name = self.format(self.counter)
        self.counter += 1

        return DataTuple(self.target_type, self.target, Path(name))

    @contextmanager
    def override(self, format: Callable[[int], str] | None = None, reset: bool = False):
        counter = self.counter
        if reset:
            self.counter = 0

        prev_format = self.format
        if format:
            self.format = format

        yield

        self.counter = counter
        self.format = prev_format


@dataclass
class ConstScoreManager:
    objective: str
    prefix: str

    constants: set[int] = field(default_factory=set, init=False)

    def format(self, value: int) -> str:
        return f"{self.prefix}{value}"

    def create(self, value: int) -> ScoreTuple:
        self.constants.add(value)

        return ScoreTuple(self.format(value), self.objective)

    __call__ = create


@dataclass
class Optimizer:
    """Handles the operation of various optimization rules.

    The `Optimizer.rule` decorates generator functions which process Operation nodes. Each rule
    must take and return a generator of Operation nodes. This usually involves a loop which
    steps through the generator and performs optimizations across the nodes.

    The optimizer runs the nodes through the rules like a conveyor belt and each rule acts as a worker.
    The rules can perform transformations, take nodes off and add different ones on, or even collect
    all nodes, and put on a completely different set of nodes. This metaphor should be considered when
    writing new rules.
    """

    temp_score: TempScoreManager
    temp_data: TempDataManager
    const_score: ConstScoreManager
    default_floating_nbt_type: str

    temp_sources: set[SourceTuple] = field(default_factory=set)
    defined_sources: set[SourceTuple] = field(default_factory=set)

    rules: list[tuple[str, Rule[IrOperation, []]]] = field(default_factory=list)

    def add_rules(self, index: int | None = None, /, **funcs: Rule[IrOperation, []]):
        """Registers new rules, also converts the decorated generator into a `SmartGenerator`"""

        if index is None:
            index = len(self.rules)

        index = min(len(self.rules), index)

        for name, f in reversed(funcs.items()):
            self.rules.insert(index, (name, f))

    @internal
    def optimize(
        self,
        nodes: Iterable[IrOperation],
        temporaries: Iterable[SourceTuple] = (),
        disable_all: bool = False,
        **rules: bool,
    ) -> tuple[Iterable[IrOperation], set[SourceTuple]]:
        """Performs the optimization by sending all nodes through the rules."""

        active_rules = {name: not disable_all for name, _ in self.rules} | rules

        with self.temp(*temporaries) as temporaries:
            for name, rule in self.rules:
                if not active_rules.get(name):
                    continue

                nodes = tuple(rule(nodes))

            return tuple(nodes), temporaries

    __call__ = optimize

    def add_temp(self, *sources: IrSource | SourceTuple):
        for source in sources:
            if isinstance(source, IrSource):
                source = source.to_tuple()

            self.temp_sources.add(source)

    @contextmanager
    def temp(self, *sources: IrSource | SourceTuple):
        prev_temp = self.temp_sources
        self.temp_sources = self.temp_sources.copy()
        self.add_temp(*sources)

        yield self.temp_sources

        self.temp_sources = prev_temp

    def is_temp(self, source: IrSource | SourceTuple):
        if isinstance(source, IrSource):
            source = source.to_tuple()

        return source in self.temp_sources

    def mark_defined(self, *sources: IrSource | SourceTuple):
        for source in sources:
            if isinstance(source, IrSource):
                source = source.to_tuple()

            self.defined_sources.add(source)

    @contextmanager
    def defined(self, *sources: IrSource | SourceTuple):
        prev_defined = set(self.defined_sources)
        self.mark_defined(*sources)

        yield

        self.defined_sources = prev_defined

    def is_defined(self, source: IrSource | SourceTuple):
        if isinstance(source, IrSource):
            source = source.to_tuple()

        return source in self.defined_sources

    def generate_score(self, temp: bool = True) -> IrScore:
        source = self.temp_score()
        if temp:
            self.add_temp(source)

        return IrScore(holder=source.holder, obj=source.obj)

    def generate_const(self, value: int) -> IrScore:
        holder, obj = self.const_score(value)
        return IrScore(holder=holder, obj=obj)

    def generate_data(self, temp: bool = True) -> IrData:
        source = self.temp_data()
        if temp:
            self.add_temp(source)

        return IrData(type=source.type, target=source.target, path=source.path)


def get_data_source_parents(node: IrData) -> tuple[IrData, ...]:
    path = cast(tuple[Accessor, ...], tuple(node.path))

    return tuple(
        replace(node, path=Path.from_accessors(path[:i]))  # type: ignore
        for i in range(len(path), 0, -1)
    )


def replace_source(
    value: IrSource, replace_map: dict[SourceTuple, IrSource]
) -> IrSource:
    if isinstance(value, IrData):
        value_path = cast(tuple[Accessor, ...], tuple(value.path))

        nodes = get_data_source_parents(value)

        for parent_node in nodes:
            replaced = False
            node = parent_node

            while new_node := replace_map.get(node.to_tuple()):
                node = cast(IrData, new_node)
                replaced = True

            if replaced:
                path = cast(
                    tuple[Accessor, ...],
                    tuple(node.path) + value_path[len(parent_node.path) :],
                )
                return replace(
                    node,
                    path=Path.from_accessors(path),  # type: ignore
                    nbt_type=value.nbt_type,
                )

        return value

    while new_node := replace_map.get(value.to_tuple()):
        value = new_node

    return value


def map_node_sources(node: Any, func: Callable[[IrSource], IrSource]) -> Any:
    if is_op(node):
        store = IrChildren(
            replace(s, value=map_node_sources(s.value, func)) for s in node.store
        )
        if isinstance(node, IrBranch):
            return replace(
                node,
                target=map_node_sources(node.target, func),
                store=store,
                children=IrChildren(map_node_sources(ch, func) for ch in node.children),
            )
        if is_unary(node):
            return replace(
                node, store=store, target=map_node_sources(node.target, func)
            )
        if is_binary(node):
            return replace(
                node,
                store=store,
                left=map_node_sources(node.left, func),
                right=map_node_sources(node.right, func),
            )

    if isinstance(node, IrUnaryCondition):
        return replace(node, target=map_node_sources(node.target, func))

    if isinstance(node, IrBinaryCondition):
        return replace(
            node,
            left=map_node_sources(node.left, func),
            right=map_node_sources(node.right, func),
        )

    if isinstance(node, IrSource):
        return func(node)

    return node


def path_accessors(path: Path) -> tuple[Accessor, ...]:
    return cast(tuple[Accessor, ...], tuple(path))


def is_path_child_of(child: Path, parent: Path) -> bool:
    child_accessors = [
        ac for ac in path_accessors(child) if not isinstance(ac, CompoundMatch)
    ]
    parent_accessors = [
        ac for ac in path_accessors(parent) if not isinstance(ac, CompoundMatch)
    ]

    if len(parent_accessors) > len(child_accessors):
        return False

    for child_accessor, parent_accessor in zip(child_accessors, parent_accessors):
        if child_accessor != parent_accessor:
            return False

    return True


def get_parent_paths(path: Path) -> Iterable[Path]:
    accessors = path_accessors(path)

    return tuple(
        Path.from_accessors(accessors[:i])  # type: ignore
        for i in range(len(accessors) - 1, 0, -1)
    )


def get_source_definitions(
    nodes: Iterable[IrOperation],
    ignore_parent: bool = False,
    ignore_children: bool = False,
) -> dict[SourceTuple, list[int]]:
    direct_definitions: dict[SourceTuple, list[int]] = {}

    for i, node in enumerate(nodes):
        for target in node.targets:
            defs = direct_definitions.setdefault(target.to_tuple(), [])
            defs.append(i)

    if ignore_parent and ignore_children:
        return direct_definitions

    definitions = {source: list(defs) for source, defs in direct_definitions.items()}

    if not ignore_children:
        # modifying child path implies in modifying the parent path
        for source in direct_definitions:
            if not isinstance(source, (DataTuple, StringDataTuple)):
                continue

            for parent_path in get_parent_paths(source.path):
                parent_source = DataTuple(source.type, source.target, parent_path)
                parent_defs = definitions.setdefault(parent_source, [])
                parent_defs.extend(
                    i for i in definitions[source] if i not in parent_defs
                )

    if not ignore_parent:
        # modifying parent path implies in modifying child paths
        for parent_source, parent_defs in direct_definitions.items():
            if not isinstance(parent_source, DataTuple):
                continue

            children_sources = [
                source
                for source in direct_definitions
                if isinstance(source, DataTuple)
                and is_path_child_of(source.path, parent_source.path)
            ]
            for source in children_sources:
                definitions[source].extend(
                    i for i in parent_defs if i not in definitions[source]
                )

    return definitions


def get_reaching_definition(
    defs: dict[SourceTuple, list[int]], source: SourceTuple, i: int
) -> int | None:
    source_defs = defs.get(source)
    if source_defs is None:
        return None

    value = bisect_left(source_defs, i) - 1
    return source_defs[value] if value >= 0 else None


def get_node_operand_dependencies(node: IrNode) -> tuple[IrSource, ...]:
    result: list[IrSource] = []

    if is_op(node):
        for operand in node.operands:
            result.extend(get_node_operand_dependencies(operand))
    elif is_unary_condition(node):
        result.extend(get_node_operand_dependencies(node.target))
    elif is_binary_condition(node):
        result.extend(get_node_operand_dependencies(node.left))
        result.extend(get_node_operand_dependencies(node.right))
    elif isinstance(node, IrSource):
        result.append(node)

    return tuple(result)


def get_node_target_dependencies(node: IrOperation) -> set[SourceTuple]:
    dependencies: set[SourceTuple] = set()

    for target in node.targets:
        if isinstance(target, IrData):
            dependencies.update(
                DataTuple(target.type, target.target, path)
                for path in get_parent_paths(target.path)
            )

    return dependencies


def get_node_dependencies(node: IrOperation) -> set[SourceTuple]:
    dependencies = set(
        source.to_tuple() for source in get_node_operand_dependencies(node)
    )
    return dependencies | get_node_target_dependencies(node)


DependencyGraph = tuple[dict[SourceTuple, set[int]], ...]


def get_dependency_graph(nodes: Iterable[IrOperation]) -> DependencyGraph:
    nodes = tuple(nodes)
    result: list[dict[SourceTuple, set[int]]] = []

    defs = get_source_definitions(nodes)
    outward_defs = get_source_definitions(nodes, ignore_children=True)

    for node_i, node in enumerate(nodes):
        dependencies: dict[SourceTuple, set[int]] = {}

        for source in get_node_operand_dependencies(node):
            source_tuple = source.to_tuple()
            dependency_set = dependencies.setdefault(source_tuple, set())
            def_i = node_i

            while True:
                def_i = get_reaching_definition(defs, source_tuple, def_i)

                if def_i is None:
                    break

                dependency_set.add(def_i)
                def_node_dependencies = get_node_dependencies(nodes[def_i])

                if source_tuple not in def_node_dependencies:
                    break

        for source_tuple in get_node_target_dependencies(node):
            def_i = get_reaching_definition(outward_defs, source_tuple, node_i)
            dependency_set = dependencies.setdefault(source_tuple, set())

            if def_i is None:
                continue

            while True:
                dependency_set.add(def_i)
                def_node_dependencies = get_node_dependencies(nodes[def_i])

                if source_tuple not in def_node_dependencies:
                    break

                def_i = get_reaching_definition(defs, source_tuple, def_i)

                if def_i is None:
                    break

        result.append(dependencies)

    return tuple(result)


Location = tuple[int, ...]


def get_source_usage(nodes: Iterable[IrOperation]) -> dict[SourceTuple, set[int]]:
    nodes = tuple(nodes)
    usage: dict[SourceTuple, set[int]] = {}

    for i, node in enumerate(nodes):
        for source in get_node_operand_dependencies(node):
            uses = usage.setdefault(source.to_tuple(), set())
            uses.add(i)

        if isinstance(node, IrBranch):
            inner_usage = get_source_usage(node.children)

            for source, inner_uses in inner_usage.items():
                if inner_uses:
                    uses = usage.setdefault(source, set())
                    uses.add(i)

    for node in nodes:
        for target in (*node.targets, *node.operands):
            if not isinstance(target, IrSource):
                continue

            uses = usage.setdefault(target.to_tuple(), set())

            if not isinstance(target, IrData):
                continue

            for path in get_parent_paths(target.path):
                parent = DataTuple(target.type, target.target, path)

                if parent_uses := usage.get(parent):
                    uses.update(parent_uses)

    for i, node in enumerate(nodes):
        for source in get_node_target_dependencies(node):
            if uses := usage.get(source):
                uses.add(i)

    return usage


def get_source_usage_of_parent(
    usage: dict[SourceTuple, set[int]], parent: SourceTuple
) -> set[int]:
    if not isinstance(parent, (DataTuple, StringDataTuple)):
        return usage.get(parent, set())

    uses: set[int] = set()

    for source in usage:
        if not isinstance(source, (DataTuple, StringDataTuple)):
            continue

        if not is_path_child_of(source.path, parent.path):
            continue

        source_uses = usage.get(source, set())
        uses.update(source_uses)

    return uses


def get_source_usage_old(nodes: Iterable[IrNode]) -> dict[SourceTuple, list[Location]]:
    map: dict[SourceTuple, list[Location]] = {}

    def add(source: Any, i: Location):
        if isinstance(source, (IrSource, SourceTuple)):
            if isinstance(source, IrSource):
                source = source.to_tuple()

            indexes = map.setdefault(source, [])
            indexes.append(i)
        elif isinstance(source, IrUnaryCondition):
            add(source.target, i)
        elif isinstance(source, IrBinaryCondition):
            add(source.left, i)
            add(source.right, i)

    for i, node in enumerate(nodes):
        if not is_op(node):
            continue

        for s in node.store:
            add(s.value, (i,))

        if is_binary(node) and node.store:
            add(node.left, (i,))

        for operand in node.operands:
            add(operand, (i,))

        if isinstance(node, IrBranch):
            children_usage = get_source_usage_old(node.children)
            for source, usage in children_usage.items():
                for u in usage:
                    add(source, (i, *u))

    return map


def data_insert_score(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    for node in nodes:
        if (
            is_binary(node, ("append", "prepend", "insert"))
            and isinstance(node.left, IrData)
            and isinstance(node.right, IrScore)
        ):
            yield replace(node, right=IrLiteral(value=Int(0)))

            if isinstance(node, IrInsert):
                index = node.index
            elif node.op == "prepend":
                index = 0
            else:
                index = -1

            element = replace(node.left, path=node.left.path[index])
            yield IrSet(left=element, right=node.right)
        else:
            yield node


def convert_data_arithmetic(opt: Optimizer, nodes: Iterable[IrOperation]):
    for node in nodes:
        if (
            is_binary(node, SCORE_OPERATIONS)
            and not is_binary(node, DATA_OPERATIONS)
            and isinstance(node.right, IrData)
        ):
            temp_score = opt.generate_score()

            yield IrCast(left=temp_score, right=node.right, cast_type=Int)
            yield replace(node, right=temp_score)

            continue

        yield node


def convert_cast(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    for node in nodes:
        if is_binary(node, "set") and (
            isinstance(node.left, IrScore)
            and isinstance(node.right, IrData)
            or isinstance(node.left, IrData)
            and isinstance(node.right, IrScore)
        ):
            cast_type = Int if isinstance(node.left, IrScore) else Any
            yield IrCast(left=node.left, right=node.right, cast_type=cast_type)
            continue

        yield node


@use_smart_generator
def noncommutative_set_collapsing(nodes: SmartGenerator[IrOperation]):
    """For noncommutative operations:
    ```
    scoreboard players operation $i1 temp = @s rx.uid
    scoreboard players operation $i1 temp -= $i0 temp
    scoreboard players operation @s rx.uid = $i1 temp
    ```

    ->
    ```
    scoreboard players operation @s rx.uid -= $i0 temp
    ```

    Examples to try:
    >>> abc["#value"] -= (1 + abc["@s"])    # doctest: +SKIP
    >>> abc["@s"] *= abc["@s"]              # doctest: +SKIP
    """

    for node in nodes:
        next_node = next(nodes, None)
        further_node = next(nodes, None)
        if (
            is_copy_op(node)
            and is_binary(next_node, SCORE_OPERATIONS)
            and is_copy_op(further_node)
            and isinstance(further_node.left, IrScore)
            and node.right == further_node.left
            and node.left == next_node.left
            and node.left == further_node.right
        ):
            yield replace(next_node, left=further_node.left, right=next_node.right)
        else:
            nodes.push(further_node)
            nodes.push(next_node)
            yield node


@use_smart_generator
def commutative_set_collapsing(
    nodes: SmartGenerator[IrOperation],
) -> Iterable[IrOperation]:
    """For commutative operations:
    ```
    scoreboard players operation $i1 temp += $i0 temp
    scoreboard players operation $i0 temp = $i1 temp
    ```
    Becomes:
    ```
    scoreboard players operation $i0 temp += $i1 temp
    ```
    """
    for node in nodes:
        next_node: Union[IrOperation, None] = next(nodes, None)

        if (
            is_binary(node, ("add", "mul"))
            and is_binary(next_node, "set")
            and isinstance(next_node.left, IrScore)
            and node.left == next_node.right
            and node.right == next_node.left
        ):
            yield replace(node, left=next_node.left, right=next_node.right)
        else:
            nodes.push(next_node)
            yield node


@use_smart_generator
def data_set_scaling(
    nodes: SmartGenerator[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    """
    Turns a multiplication/division of a temp score followed by a
    data set operation into a single set operation with a scale argument.
    ```
    scoreboard players operation $i0 temp /= $100 const
    execute store result storage demo out int 1 run scoreboard players get $i0 temp
    ```
    ->
    ```
    execute store result storage demo out float 0.01 run scoreboard players get $i0 temp
    ```

    Also works with all data operations such as append, prepend and insert:
    ```
    scoreboard players operation $i1 bolt.expr.temp *= $100 bolt.expr.const
    data modify storage demo list append value 0
    execute store result storage demo list[-1] int 1 run scoreboard players get $i1 bolt.expr.temp
    ```
    ->
    ```
    data modify storage demo list append value 0
    execute store result storage demo list[-1] int 100 run scoreboard players get $i1 bolt.expr.temp

    ```
    Examples to try:
    >>> temp.out = (obj["$value"] + 1) * 10
    >>> temp.percent = (obj["$stack"] * 100) / 64
    """
    for node in nodes:
        operation_node = None
        next_node = next(nodes, None)
        # skip data operation node (just so we can get to
        # the actual Set node)
        if is_binary(next_node, ("insert", "append", "prepend")):
            operation_node = next_node
            next_node = next(nodes, None)
        if (
            is_binary(node, ("mul", "div"))
            and isinstance(next_node, IrCast)
            and isinstance(node.right, IrLiteral)
            and isinstance(node.right.value, Numeric)
            and isinstance(next_node.left, IrData)
            and node.left == next_node.right
        ):
            scale = float(node.right.value)

            if scale.is_integer():
                scale = int(scale)

            number_type = next_node.cast_type

            if is_binary(node, "div"):
                scale = 1 / scale

                if number_type is Any:
                    number_type = literal_types[opt.default_floating_nbt_type]

            out = IrCast(
                left=next_node.left, right=node.left, cast_type=number_type, scale=scale
            )

            if operation_node:
                yield operation_node  # yield the data operation node back in

            yield out
        else:
            nodes.push(next_node)
            nodes.push(operation_node)
            yield node


@use_smart_generator
def data_get_scaling(nodes: SmartGenerator[IrOperation]) -> Iterable[IrOperation]:
    """
    ````
    execute store result score $i0 temp run data get storage demo value 1
    scoreboard players operation $i0 temp *= $100 const
    ```
    ->
    ```
    execute store result score $i0 temp run data get storage demo value 100
    ```
    Examples to try:
    >>> obj["#offset"] = (player.Motion[0] * 100) - obj["#x"]
    """
    for node in nodes:
        next_node = next(nodes, None)
        if (
            isinstance(node, IrCast)
            and is_binary(next_node, ("mul", "div"))
            and isinstance(node.right, IrData)
            and isinstance(next_node.right, IrLiteral)
            and isinstance(next_node.right.value, Numeric)
            and node.left == next_node.left
        ):
            scale = float(next_node.right.value)

            if scale.is_integer():
                scale = int(scale)

            if is_binary(next_node, "div"):
                scale = 1 / scale

            yield IrCast(left=node.left, right=node.right, scale=node.scale * scale)
        else:
            nodes.push(next_node)
            yield node


def multiply_divide_by_fraction(nodes: Iterable[IrOperation]):
    for node in nodes:
        if (
            is_binary(node, ("mul", "div"))
            and isinstance(node.right, IrLiteral)
            and isinstance(node.right.value, (Float, Double))
        ):
            value = Fraction(node.right.value).limit_denominator()

            if is_binary(node, "div"):
                value = 1 / value

            numerator = IrLiteral(value=Int(value.numerator))
            denominator = IrLiteral(value=Int(value.denominator))

            yield IrBinary(op="mul", left=node.left, right=numerator)
            yield IrBinary(op="div", left=node.left, right=denominator)
        else:
            yield node


def multiply_divide_by_one_removal(nodes: Iterable[IrOperation]):
    for node in nodes:
        if not (
            is_binary(node, ("mul", "div"))
            and isinstance(node.right, IrLiteral)
            and node.right.value == 1
        ):
            yield node


def add_subtract_by_zero_removal(nodes: Iterable[IrOperation]):
    for node in nodes:
        if not (
            is_binary(node, ("add", "sub"))
            and isinstance(node.right, IrLiteral)
            and node.right.value == 0
        ):
            yield node


def discard_casting(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    for node in nodes:
        if isinstance(node, IrCast) and is_copy_op(node):
            yield IrSet(left=node.left, right=node.right)
        else:
            yield node


def discard_non_numerical_casting(
    nodes: Iterable[IrOperation],
) -> Iterable[IrOperation]:
    for node in nodes:
        if (
            isinstance(node, IrCast)
            and isinstance(node.left, IrData)
            and isinstance(node.right, IrData)
            and node.scale == 1
            and not is_numeric_type(unwrap_optional_type(node.cast_type))
        ):
            yield IrSet(left=node.left, right=node.right)
        else:
            yield node


def set_to_self_removal(nodes: Iterable[IrOperation]):
    """Removes Set operations that have the same former and latter source.
    Should run after "output_score_replacement" is applied to clean up
    all the reduntant Sets created by the previous rule.
    """
    for node in nodes:
        if is_copy_op(node):
            if (
                not isinstance(node.right, IrSource)
                or type(node.left) is not type(node.right)
                or node.left.to_tuple() != node.right.to_tuple()
            ):
                yield node
        else:
            yield node


def literal_to_constant_replacement(
    opt: Optimizer,
    nodes: Iterable[IrOperation],
) -> Iterable[IrOperation]:
    for node in nodes:
        if (
            is_binary(node, ("mul", "div", "min", "max", "mod"))
            and isinstance(node.right, IrLiteral)
            and isinstance(node.right.value, Numeric)
        ):
            value = int(node.right.value)
            constant = opt.generate_const(value)

            yield replace(node, right=constant)
        else:
            yield node


def set_and_get_cleanup(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    """
    Removes unnecessary temp vars possibly originated from previous
    optimizations.
    ```
    scoreboard players operation $i0 bolt.expr.temp = $value obj
    execute store result storage demo out float 0.01 run scoreboard players get $i0 bolt.expr.temp
    ```
    ->
    ```
    execute store result storage demo out float 0.01 run scoreboard players get $value obj
    ```
    Examples to try:
    >>> temp.out = obj["$value"] / 100
    >>> temp.out = temp.value * 100 / 100
    """

    nodes = tuple(nodes)
    usage = get_source_usage(nodes)
    definitions = get_source_definitions(nodes)

    operations: list[IrOperation] = []

    for node_i, node in enumerate(nodes):
        operands: list[OperandType] = []

        for operand in node.operands:
            operands.append(operand)

            if not isinstance(operand, IrSource):
                continue

            source = operand.to_tuple()
            def_i = get_reaching_definition(definitions, source, node_i)

            if def_i is None:
                continue

            def_node = nodes[def_i]

            if (
                not isinstance(def_node, (IrSet, IrCast))
                or not isinstance(def_node.right, IrSource)
                or def_node.left.to_tuple() != source
                or operand in node.targets
            ):
                continue

            if isinstance(def_node, IrCast) and (
                def_node.scale != 1.0
                or not isinstance(def_node.left, IrScore)
                or not isinstance(node, (IrSet, IrCast))
            ):
                continue

            if any(
                def_i < use_i < node_i
                for use_i in get_source_usage_of_parent(usage, source)
            ):
                continue

            operands[-1] = def_node.right

        operations.append(replace_operation(node, operands=operands))

    yield from convert_cast(operations)


def rename_temp_scores(
    opt: Optimizer,
    nodes: Iterable[IrOperation],
) -> Iterable[IrOperation]:
    nodes = tuple(nodes)

    ignored_sources: set[IrSource] = set()

    defs = get_source_definitions(nodes, ignore_parent=True, ignore_children=True)
    for i, node in enumerate(nodes):
        for source in get_node_operand_dependencies(node):
            source_defs = defs.get(source.to_tuple(), [])

            if not any(def_i < i for def_i in source_defs):
                ignored_sources.add(source)

    with (
        opt.temp_score.override(
            format=lambda n: f"{opt.temp_score.prefix}i{n}", reset=True
        ),
        opt.temp_data.override(format=lambda n: f"i{n}", reset=True),
    ):
        replace_map: dict[SourceTuple, IrSource] = {}

        def map_source(node: IrSource) -> IrSource:
            if node in ignored_sources:
                return node

            replaced_node = replace_source(node, replace_map)

            if replaced_node != node:
                return replaced_node

            if not opt.is_temp(node):
                return node

            node_tuple = node.to_tuple()

            if isinstance(node, IrScore):
                replace_map[node_tuple] = opt.generate_score()
            elif isinstance(node, IrData):
                replace_map[node_tuple] = opt.generate_data()
            else:
                replace_map[node_tuple] = node

            return replace_map[node_tuple]

        nodes = [map_node_sources(node, map_source) for node in nodes]

    yield from nodes


def data_string_propagation(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    all_nodes = tuple(nodes)
    defs = get_source_definitions(all_nodes)

    for i, node in enumerate(all_nodes):
        if not is_binary(
            node, ("append", "prepend", "merge", "insert", "set")
        ) or not isinstance(node.right, IrSource):
            yield node
            continue

        cond_def_i = get_reaching_definition(defs, node.right.to_tuple(), i)
        if cond_def_i is None:
            yield node
            continue

        cond_def = all_nodes[cond_def_i]

        if (
            is_binary(cond_def, "set")
            and isinstance(cond_def.right, IrDataString)
            and cond_def.left.to_tuple() == node.right.to_tuple()
        ):
            yield replace(node, right=cond_def.right)
            continue

        yield node


def negate_condition(node: IrCondition):
    if is_unary_condition(node, "not"):
        return IrUnaryCondition(op="truthy", target=node.target)

    return replace(node, negated=not node.negated)


def boolean_condition_propagation(
    nodes: Iterable[IrOperation],
) -> Iterable[IrOperation]:
    all_nodes = list(nodes)
    defs = get_source_definitions(all_nodes)

    for i, node in enumerate(all_nodes):
        if is_binary(node, "set") and is_unary_condition(node.right, "boolean"):
            bool_cond = node.right

            cond_def_i = get_reaching_definition(defs, bool_cond.target.to_tuple(), i)
            if cond_def_i is None:
                continue

            cond_node = all_nodes[cond_def_i]

            if (
                is_binary(cond_node, "set")
                and is_condition(cond_node.right)
                and cond_node.left.to_tuple() == bool_cond.target.to_tuple()
            ):
                right = (
                    negate_condition(cond_node.right)
                    if bool_cond.negated
                    else cond_node.right
                )
                all_nodes[i] = replace(node, right=right)

    yield from all_nodes


def branch_condition_propagation(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    all_nodes = tuple(nodes)
    defs = get_source_definitions(all_nodes)

    for i, node in enumerate(all_nodes):
        if not is_unary(node, "branch") or not isinstance(node.target, IrSource):
            yield node
            continue

        cond_def_i = get_reaching_definition(defs, node.target.to_tuple(), i)
        if cond_def_i is None:
            yield node
            continue

        cond_def = all_nodes[cond_def_i]

        if (
            is_binary(cond_def, "set")
            and is_condition(cond_def.right)
            and node.target.to_tuple() == cond_def.left.to_tuple()
        ):
            yield replace(node, target=cond_def.right)
            continue

        yield node


def deadcode_elimination(
    nodes: Iterable[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    all_nodes = tuple(nodes)
    usage = get_source_usage_old(all_nodes)

    for node_i, node in enumerate(all_nodes):
        if isinstance(node, IrBranch):
            yield node
            continue

        store: list[IrStore] = []

        for store_el in node.store:
            source = store_el.value

            if not opt.is_temp(source) or any(
                use_i > (node_i,) for use_i in usage.get(source.to_tuple(), [])
            ):
                store.append(store_el)

        if store != node.store:
            node = replace(node, store=IrChildren(store))

        for target in node.targets:
            if not opt.is_temp(target):
                yield node
                break

            target_usage = usage.get(target.to_tuple(), [])

            if any(use_i > (node_i,) for use_i in target_usage):
                yield node
                break


def convert_data_order_operation(
    nodes: Iterable[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    result: list[IrOperation] = []

    OPERATIONS = (
        "less_than",
        "less_than_or_equal_to",
        "greater_than",
        "greater_than_or_equal_to",
        "equal",
    )

    def replace_operand(node: Any, in_condition: bool = False) -> Any:
        if is_binary_condition(node) and node.op in OPERATIONS:
            return replace(
                node,
                left=replace_operand(node.left, in_condition=True),
                right=replace_operand(node.right, in_condition=True),
            )

        if is_unary_condition(node) and node.op in OPERATIONS:
            return replace(node, target=replace_operand(node.target, in_condition=True))

        if isinstance(node, IrData) and in_condition:
            score = opt.generate_score()
            result.append(IrCast(left=score, right=node, cast_type=Int))
            return score

        return node

    for node in nodes:
        if is_unary(node):
            node = replace(node, target=replace_operand(node.target))
        elif is_binary(node):
            node = replace(
                node, left=replace_operand(node.left), right=replace_operand(node.right)
            )

        result.append(node)

    yield from result


def compound_match_data_compare(
    nodes: Iterable[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    for node in nodes:
        operands: list[OperandType] = []

        for operand in node.operands:
            if (
                is_binary_condition(operand, "equal")
                and isinstance(operand.left, IrData)
                and isinstance(operand.right, IrLiteral)
            ):
                value = operand.right.value

                if not isinstance(value, (Numeric, String)):
                    operands.append(operand)
                    continue

                source = operand.left
                negated = operand.negated
                path = cast(tuple[Accessor, ...], tuple(source.path))
                accessor = path[-1] if len(path) else None
                match = None

                if isinstance(accessor, NamedKey):
                    subpath = path[:-1]
                    match = CompoundMatch(Compound({accessor.key: value}))
                else:
                    temp = opt.generate_data(temp=False)

                    yield IrUnary(op="remove", target=temp)
                    yield IrSet(left=temp, right=source)

                    accessor = cast(NamedKey, tuple(temp.path)[-1])
                    source = temp
                    subpath = ()
                    match = CompoundMatch(Compound({accessor.key: value}))

                new_path = Path.from_accessors((*subpath, match))  # type: ignore
                new_source = replace(source, path=new_path)
                operand = IrUnaryCondition(
                    op="boolean", target=new_source, negated=negated
                )

            operands.append(operand)

        yield replace_operation(node, operands=operands)


def store_set_data_compare(
    nodes: Iterable[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    for node in nodes:
        operands: list[OperandType] = []

        for operand in node.operands:
            if is_binary_condition(operand, "equal"):
                data = operand.left
                value = operand.right

                if not isinstance(data, IrData):
                    data, value = value, data

                if not isinstance(data, IrData) or not isinstance(
                    value, (IrData, IrLiteral)
                ):
                    operands.append(operand)
                    continue

                result = opt.generate_score()
                cmp = opt.generate_data()

                opt.mark_defined(result)
                yield IrSet(left=result, right=IrLiteral(value=Int(1)))
                yield IrSet(left=cmp, right=data)

                store = IrStore(type=StoreType.success, value=result)
                set_op = IrSet(store=IrChildren((store,)), left=cmp, right=value)

                if isinstance(value, IrData):
                    inner_op = IrBranch(
                        target=IrUnaryCondition(op="boolean", target=value),
                        children=IrChildren((set_op,)),
                    )
                else:
                    inner_op = set_op

                yield IrBranch(
                    target=IrUnaryCondition(op="boolean", target=data),
                    children=IrChildren((inner_op,)),
                )

                operand = IrUnaryCondition(op="boolean", target=result, negated=True)

            operands.append(operand)

        yield replace_operation(node, operands=operands)


def init_score_boolean_result(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    all_nodes = tuple(nodes)
    defs = get_source_definitions(all_nodes)

    for i, node in enumerate(all_nodes):
        if (
            is_binary(node, "set")
            and is_unary_condition(node.right, "boolean")
            and isinstance(node.right.target, IrScore)
            and not any(x < i for x in defs.get(node.left.to_tuple(), []))
        ):
            yield IrSet(left=node.left, right=IrLiteral(value=Int(0)))

        yield node


Op = TypeVar("Op", bound=IrOperation)


def replace_operation(
    node: Op, operands: Iterable[OperandType] | None = None, **kwargs: Any
) -> Op:
    if operands is None:
        operands = node.operands
    else:
        operands = tuple(operands)

    if is_unary(node):
        return replace(node, target=operands[0], **kwargs)

    if is_binary(node):
        if len(operands) == 1:
            return replace(node, right=operands[0], **kwargs)

        return replace(node, left=operands[0], right=operands[1], **kwargs)

    return replace(node, **kwargs)


def convert_defined_boolean_condition(
    nodes: Iterable[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    for node in nodes:
        operands: list[OperandType] = []

        for operand in node.operands:
            if is_unary(node, "branch") and isinstance(operand, IrSource):
                operand = IrUnaryCondition(op="boolean", target=operand)

            if (
                is_unary_condition(operand, "boolean")
                and isinstance(operand.target, IrScore)
                and opt.is_defined(operand.target)
            ):
                negated = not operand.negated
                target = operand.target
                zero = IrLiteral(value=Int(0))

                operand = IrBinaryCondition(
                    op="equal", left=target, right=zero, negated=negated
                )

            operands.append(operand)

        yield replace_operation(node, operands=tuple(operands))


def store_result_inlining(nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:
    nodes = tuple(nodes)

    defs = get_source_definitions(nodes)

    stores: dict[int, list[IrStore]] = {}
    removed: set[int] = set()

    for i, node in enumerate(nodes):
        if not isinstance(node, IrCast) or not isinstance(node.right, IrSource):
            continue

        source_def_i = get_reaching_definition(defs, node.right.to_tuple(), i)

        if source_def_i is None or source_def_i != (i - 1):
            continue

        def_node = nodes[source_def_i]

        if isinstance(def_node, (IrSet, IrCast)):
            continue

        store = IrStore(
            type=StoreType.result,
            value=node.left,
            cast_type=node.cast_type,
            scale=node.scale,
        )
        source_stores = stores.setdefault(source_def_i, [])
        source_stores.append(store)
        removed.add(i)

    for i, node in enumerate(nodes):
        if store := stores.get(i):
            node = replace(node, store=IrChildren((*node.store, *store)))

        if i not in removed:
            yield node


def get_default_value(type: NbtType) -> NbtValue:
    type = unwrap_optional_type(type)

    if is_numeric_type(type):
        return type(0)

    if is_string_type(type):
        return String("")

    if is_list_type(type):
        return List([])

    if is_array_type(type):
        return type([])

    if is_compound_type(type):
        return Compound({})

    return Int(0)


def traverse_composite_literal(
    value: CompositeNbtValue,
    result: IrData,
    type: NbtType,
    ctx: Context,
    operations: list[IrOperation],
) -> NbtValue:
    literal: Any

    if isinstance(value, dict):
        value = cast(dict[str, CompositeNbtValue | IrSource], value)

        literal = {}

        for key, val in value.items():
            val_result = replace(result, path=result.path[key])

            if isinstance(val, IrSource):
                accessors = cast(tuple[Accessor, ...], tuple(val_result.path))
                cast_type = access_type_by_path(type, accessors[1:], ctx)

                if cast_type in (None, Any):
                    cast_type = val.nbt_type if isinstance(val, IrData) else Any

                operations.append(
                    IrCast(left=val_result, right=val, cast_type=cast_type or Any)
                )
                nbt_val = get_default_value(cast_type)
            else:
                nbt_val = traverse_composite_literal(
                    val, val_result, type, ctx, operations
                )

            literal[key] = nbt_val

    elif isinstance(value, list):
        value = cast(list[CompositeNbtValue | IrSource], value)

        literal = []

        accessors = cast(tuple[Accessor, ...], tuple(result.path[0]))
        cast_type = access_type_by_path(type, accessors[1:], ctx)

        if cast_type in (None, Any):
            literals = [el for el in value if not isinstance(el, IrSource)]

            if literals:
                cast_type = infer_type(literals[0], shallow=True)

            if cast_type in (None, Any):
                data_source_types = [
                    el.nbt_type
                    for el in value
                    if isinstance(el, IrData) and el.nbt_type is not Any
                ]
                cast_type = data_source_types[0] if data_source_types else Int

        for i, val in enumerate(value):
            val_result = replace(result, path=result.path[i])

            if isinstance(val, IrSource):
                operations.append(
                    IrCast(left=val_result, right=val, cast_type=cast_type)
                )
                nbt_val = get_default_value(cast_type)
            else:
                nbt_val = traverse_composite_literal(
                    val, val_result, type, ctx, operations
                )

            literal.append(nbt_val)
    else:
        literal = value

    nbt = convert_tag(literal)

    if nbt is None:
        raise ValueError("Invalid nbt.")

    return nbt


def composite_literal_expansion(
    nodes: Iterable[IrOperation], opt: Optimizer, ctx: Context
) -> Iterable[IrOperation]:
    for node in nodes:
        if not any(
            isinstance(operand, IrCompositeLiteral) for operand in node.operands
        ):
            yield node
            continue

        operands: list[OperandType] = []

        for operand in node.operands:
            if isinstance(operand, IrCompositeLiteral):
                result = opt.generate_data()

                if isinstance(node, IrCast):
                    type = node.cast_type
                elif is_binary(node, ("append", "prepend", "insert")) and isinstance(
                    node.left, IrData
                ):
                    type = (
                        access_type_by_path(node.left.nbt_type, (ListIndex(0),)) or Any
                    )
                else:
                    type = Any

                operations: list[IrOperation] = []
                nbt_value = traverse_composite_literal(
                    operand.value, result, type, ctx, operations
                )

                yield IrCast(
                    left=result, right=IrLiteral(value=nbt_value), cast_type=type
                )
                yield from operations

                operand = result

            operands.append(operand)

        yield replace_operation(node, operands)


def source_copy_elision(
    nodes: Iterable[IrOperation], opt: Optimizer
) -> Iterable[IrOperation]:
    nodes = list(nodes)
    active = True

    while active:
        active = False

        definitions = get_source_definitions(nodes)
        dependency = get_dependency_graph(nodes)
        usage = get_source_usage(nodes)

        for node_i in range(len(nodes)):
            node = nodes[node_i]
            if not (
                is_binary(node, "set")
                and isinstance(node.right, IrSource)
                and opt.is_temp(node.right)
            ):
                continue

            target = node.left.to_tuple()
            target_source = node.left
            source = node.right.to_tuple()

            if any(
                use_i > node_i for use_i in get_source_usage_of_parent(usage, source)
            ):
                continue

            node_dependencies = dependency[node_i]
            deps = sorted(node_dependencies.get(source, set()))

            if not deps:
                continue

            if any(
                deps[0] < target_use_i < node_i
                for target_use_i in get_source_usage_of_parent(usage, target)
            ):
                continue

            conflicting_defs = sorted(
                def_i
                for def_i in definitions.get(target, [])
                if deps[0] < def_i < node_i
            )
            if conflicting_defs:
                def_i = conflicting_defs[0]

                def_deps = tuple(
                    s for source_deps in dependency[def_i].values() for s in source_deps
                )
                if any(deps[0] < dep_i for dep_i in def_deps):
                    continue

                def_node = nodes.pop(def_i)
                nodes.insert(deps[0], def_node)

                active = True
                break

            for i in deps:
                nodes[i] = map_node_sources(
                    nodes[i], lambda s: replace_source(s, {source: target_source})
                )

            nodes.pop(node_i)

            active = True
            break

    return nodes



from contextlib import nullcontext, suppress
from dataclasses import dataclass
from typing import (
    Any,
    Iterable,
    TypedDict,
    get_args,
    get_type_hints,
)

from beet import Context
from bolt.utils import internal
from mecha import Visitor, rule
from nbtlib import (
    Array,
    Byte,
    Double,
    Float,
    Int,
    ListIndex,
    Long,
    NamedKey,
    Numeric,
    Short,
)  # type: ignore

from .exceptions import TypeCheckDiagnostic, TypeCheckError, get_exception_chain
from .optimizer import (
    IrBinary,
    IrCast,
    IrData,
    IrLiteral,
    IrOperation,
    IrScore,
    IrSource,
)
from .typing import (
    NbtType,
    NumericNbtValue,
    access_type,
    convert_type,
    format_type,
    get_dict_fields,
    infer_type,
    is_alias,
    is_array_type,
    is_compound_alias,
    is_fixed_compound,
    is_list_type,
    is_numeric_type,
    is_optional,
    is_type,
    is_typeddict_guard,
    is_union,
    unwrap_optional_type,
)
from .utils import get_globals

__all__ = [
    "TypeCheckFlags",
    "check_union_type",
    "check_typeddict_type",
    "check_expandable_compound_type",
    "check_list_type",
    "check_numeric_type",
    "check_type",
    "TypeChecker",
]


NUMERIC_ORDER = (Byte, Short, Int, Long, Float, Double)


class TypeCheckFlags(TypedDict, total=False):
    suppress: bool
    numeric_match: bool
    ignore_missing_keys: bool


def check_union_type(
    write: NbtType,
    read: NbtType,
    ctx: Context | None = None,
    **flags: Any,
) -> bool:
    if is_union(read):
        return all(check_type(write, r, ctx, **flags) for r in get_args(read))

    if is_union(write):
        flags = {**flags, "suppress": True}

        if not any(check_type(w, read, ctx, **flags) for w in get_args(write)):
            raise TypeCheckError(
                f'"{format_type(read)}" is not compatible with "{format_type(write)}".'
            )

        return True

    return False


def check_typeddict_type(
    write: type[TypedDict],
    read: NbtType,
    ctx: Context | None = None,
    **flags: Any,
) -> bool:
    if not is_fixed_compound(read):
        raise TypeCheckError(
            f'"{format_type(read)}" is not a compound type with fixed keys and is not compatible with "{format_type(write)}".'
        )

    if write is read:
        return True

    new_flags = {"numeric_match": True, **flags, "ignore_missing_keys": False}

    write_annotations = get_dict_fields(write, ctx)
    read_annotations = get_dict_fields(read, ctx)

    opt_keys = set(write.__optional_keys__)
    opt_keys.update(
        name for name, type in write_annotations.items() if is_optional(type)
    )

    for key, key_type in write_annotations.items():
        if key not in read_annotations:
            if key in opt_keys or flags.get("ignore_missing_keys"):
                continue

            raise TypeCheckError(
                f'"{format_type(read)}" is missing required key "{key}" of type "{format_type(key_type)}".'
            )

        try:
            if not check_type(
                write_annotations[key], read_annotations[key], ctx, **new_flags
            ):
                return False
        except TypeCheckError as cause_exc:
            exc = TypeCheckError(
                f'Key "{key}" of "{format_type(read)}" is incompatible with "{format_type(write)}":'
            )
            raise exc from cause_exc

    for key in read_annotations:
        if key not in write_annotations:
            raise TypeCheckError(
                f'"{format_type(read)}" has extra key "{key}" not present in "{format_type(write)}".'
            )

    return True


def check_expandable_compound_type(
    write: type[dict[str, Any]],
    read: NbtType,
    ctx: Context | None = None,
    **flags: Any,
) -> bool:
    flags = {"numeric_match": True, **flags, "ignore_missing_keys": False}

    child_type = access_type(write, NamedKey(""))

    if is_alias(read, dict):
        read_child_type = access_type(read, NamedKey(""))

        try:
            return check_type(child_type, read_child_type, ctx, **flags)
        except TypeCheckError as cause_exc:
            exc = TypeCheckError(
                f'"{format_type(read)}" and "{format_type(write)}" have incompatible key types:'
            )
            raise exc from cause_exc

    if is_fixed_compound(read):
        if isinstance(read, dict):
            read_annotations = read
        else:
            read_annotations = get_type_hints(read, get_globals(read, ctx))

        for key, type in read_annotations.items():
            type = unwrap_optional_type(type)

            try:
                if not check_type(child_type, type, ctx, **flags):
                    return False
            except TypeCheckError as cause_exc:
                exc = TypeCheckError(
                    f'"{format_type(read)}" key "{key}" is not valid key of "{format_type(write)}":'
                )
                raise exc from cause_exc

        return True

    raise TypeCheckError(
        f'"{format_type(read)}" is not a compound type and is not compatible with "{format_type(write)}".'
    )


def check_list_type(
    write: type[list[Any] | Array],
    read: NbtType,
    ctx: Context | None = None,
    **flags: Any,
) -> bool:
    flags = {**flags, "numeric_match": True, "ignore_missing_keys": False}

    subtype = access_type(write, ListIndex(None))
    read_subtype = access_type(read, ListIndex(None))

    if read_subtype is None:
        raise TypeCheckError(
            f'"{format_type(read)}" is not a list/array type and is not compatible with "{format_type(write)}".'
        )

    try:
        return check_type(subtype, read_subtype, ctx, **flags)
    except TypeCheckError as cause_exc:
        exc = TypeCheckError(
            f'Elements of "{format_type(read)}" and "{format_type(write)}" are not compatible:'
        )
        raise exc from cause_exc


def check_numeric_type(write: type[NumericNbtValue], read: Any, **flags: Any) -> bool:
    if not issubclass(read, Numeric):
        raise TypeCheckError(
            f'"{format_type(read)}" is not a numeric type and is not compatible with "{format_type(write)}".'
        )

    numeric_match = flags.get("numeric_match", False)

    write_order = NUMERIC_ORDER.index(write)
    read_order = NUMERIC_ORDER.index(read)

    if numeric_match and write_order < read_order:
        raise TypeCheckError(
            f'"{format_type(read)}" cannot be implicitly narrowed to "{format_type(write)}".'
        )

    if numeric_match and read_order < write_order:
        raise TypeCheckError(
            f'"{format_type(read)}" cannot be implicitly converted to "{format_type(write)}".'
        )

    return True


def check_type(
    write: NbtType | None,
    read: NbtType,
    ctx: Context | None = None,
    **flags: Any,
) -> bool:
    """Checks if the type `read` is compatible with the type `write`."""

    context = suppress(TypeCheckError) if flags.get("suppress") else nullcontext()
    flags = {**flags, "suppress": False}

    result = False

    with context:
        if write is None:
            return False

        if write is Any or read is Any:
            return True

        if is_union(write) or is_union(read):
            return check_union_type(write, read, ctx, **flags)

        if is_typeddict_guard(write):
            return check_typeddict_type(write, read, ctx, **flags)

        if is_compound_alias(write):
            return check_expandable_compound_type(write, read, ctx, **flags)

        if is_list_type(write) or is_array_type(write):
            return check_list_type(write, read, ctx, **flags)

        if is_numeric_type(write):
            return check_numeric_type(write, read, **flags)

        if convert_type(write) != convert_type(read):
            raise TypeCheckError(
                f'"{format_type(read)}" does not match "{format_type(write)}".'
            )

        result = True

    return result


@dataclass(eq=False, kw_only=True)
class TypeChecker(Visitor):
    ctx: Context | None

    @internal
    def __call__(self, nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:  # type: ignore
        for node in nodes:
            exc = self.invoke(node)

            if exc is not None:
                raise exc

            yield node

    def get_type(self, node: IrSource | IrLiteral) -> NbtType:
        if isinstance(node, IrData):
            return node.nbt_type
        elif isinstance(node, IrScore):
            return Int
        elif isinstance(node, IrLiteral):
            t = infer_type(node.value)
            return t if t is not None else Any

        return Any

    def check_type(
        self,
        former: IrSource,
        latter: IrSource | IrLiteral,
        write: NbtType | None = None,
        read: NbtType | None = None,
        **flags: Any,
    ) -> tuple[bool, tuple[BaseException, ...]]:
        if write is None:
            write = self.get_type(former)
        if read is None:
            read = self.get_type(latter)

        match = False
        errors: tuple[BaseException, ...] = ()
        # require numeric types to match if right side is literal, that means
        # it fails only if a literal failed to be casted to the right type.
        flags = {"numeric_match": isinstance(latter, IrLiteral), **flags}

        try:
            match = check_type(write, read, self.ctx, **flags)
        except TypeCheckError as exc:
            errors = get_exception_chain(exc)

        return match, errors

    def format_operand(self, value: IrSource | IrLiteral, read: NbtType | None = None):
        if read is None:
            read = self.get_type(value)

        if isinstance(value, IrData):
            return f'{value.type} "{value.target} {value.path}" of type "{format_type(read)}"'
        if isinstance(value, IrScore):
            return f'score "{value.holder} {value.obj}"'
        return f'value of type "{format_type(read)}"'

    def format_errors(self, errors: Any) -> str:
        return "".join("\n  - " + str(e) for e in errors)

    def create_diagnostic(
        self,
        pattern: str,
        *args: NbtType
        | tuple[IrSource | IrLiteral, NbtType]
        | tuple[Any, ...]
        | IrSource
        | IrLiteral
        | Any,
    ) -> TypeCheckDiagnostic:
        values: list[str] = []

        for arg in args:
            if is_type(arg):
                value = format_type(arg)
            elif isinstance(arg, tuple):
                if len(arg) == 2 and isinstance(arg[0], (IrSource, IrLiteral)):
                    value = self.format_operand(arg[0], arg[1])
                elif len(arg) and isinstance(arg[0], TypeCheckError):
                    value = self.format_errors(arg)
                else:
                    value = "".join(str(el) for el in arg)
            elif isinstance(arg, (IrSource, IrLiteral)):
                value = self.format_operand(arg)
            else:
                value = str(arg)

            values.append(value)

        fmt = pattern % tuple(values)
        msg = fmt[0].capitalize() + fmt[1:]
        return TypeCheckDiagnostic(msg)

    @rule(IrOperation)
    def fallback(self, node: IrOperation) -> None: ...

    @rule(IrCast)
    def set(self, node: IrCast) -> TypeCheckDiagnostic | None:
        write = node.cast_type
        matched, errors = self.check_type(node.left, node.right, write)

        if not matched:
            return self.create_diagnostic(
                "%s cannot be assigned to %s: %s",
                node.right,
                (node.left, write),
                errors,
            )

    @rule(IrBinary, op="append")
    @rule(IrBinary, op="prepend")
    @rule(IrBinary, op="insert")
    def insert(self, node: IrBinary) -> TypeCheckDiagnostic | None:
        write = access_type(self.get_type(node.left), ListIndex(None))

        if write is None:
            return self.create_diagnostic(
                '%s is not a list/array and does not support "%s".', node.left, node.op
            )

        matched, errors = self.check_type(
            node.left, node.right, write, numeric_match=True
        )

        if not matched:
            if node.op == "append":
                msg = "appended to"
            elif node.op == "prepend":
                msg = "prepended to"
            else:
                msg = "inserted into"

            return self.create_diagnostic(
                "%s cannot be %s %s: %s",
                node.right,
                msg,
                node.left,
                errors,
            )

    @rule(IrBinary, op="merge")
    def merge(self, node: IrBinary) -> TypeCheckDiagnostic | None:
        matched, errors = self.check_type(
            node.left, node.right, ignore_missing_keys=True
        )

        if not matched:
            return self.create_diagnostic(
                "%s cannot be merged with %s: %s",
                node.right,
                node.left,
                errors,
            )



from dataclasses import dataclass, field
from typing import Any, Generator, Iterable, cast

from mecha import AstChildren, AstCommand, AstRoot, Mecha, Visitor, rule
from mecha.utils import number_to_string
from nbtlib import Byte, Double, Float, Int, Long, Short  # type: ignore

from .optimizer import (
    IrBinary,
    IrBinaryCondition,
    IrBoolScore,
    IrBranch,
    IrCast,
    IrChildren,
    IrCondition,
    IrData,
    IrDataString,
    IrInsert,
    IrLiteral,
    IrNode,
    IrOperation,
    IrRaw,
    IrScore,
    IrSet,
    IrSource,
    IrStore,
    IrUnary,
    IrUnaryCondition,
)
from .typing import (
    NBT_TYPE_STRING,
    NbtTypeString,
    NumericNbtValue,
    unwrap_optional_type,
)
from .utils import insert_nested_commands, type_name

__all__ = [
    "InvalidOperand",
    "AstConverter",
]


class InvalidOperand(Exception):
    def __init__(self, op: str, *operands: Any):
        fmt = ", ".join(f"'{type_name(operand)}'" for operand in operands)
        super().__init__(f"Invalid operand(s) for '{op}' operation: {fmt}.")


@dataclass(kw_only=True)
class AstConverter(Visitor):
    default_nbt_type: NbtTypeString

    mc: Mecha
    result: list[AstCommand] = field(default_factory=list)

    def __call__(self, nodes: Iterable[IrOperation]) -> AstChildren[AstCommand]:  # type: ignore
        prev_result = self.result
        self.result = []

        for node in nodes:
            self.invoke(node)

        result = AstChildren(self.result)
        self.result = prev_result
        return result

    def add_result(
        self,
        cmd: str,
        store: IrChildren[IrStore] | None = None,
        children: AstChildren[AstCommand] | None = None,
    ) -> None:
        if store:
            prefix = tuple(self.invoke(s) for s in store)
            cmd = " ".join((*prefix, cmd))

        node = self.mc.parse(cmd, using="command")

        if children:
            node = insert_nested_commands(node, AstRoot(commands=children))

        self.result.append(node)

    @rule(IrNode)
    def fallback(self, node: IrNode):
        raise TypeError(f"Could not convert object '{node}' to AST.")

    @rule(IrRaw)
    def raw(self, node: IrRaw[AstCommand]):
        self.result.append(node.node)

    @rule(IrScore)
    def score(self, node: IrScore) -> str:
        return f"{node.holder} {node.obj}"

    @rule(IrData)
    def data(self, node: IrData) -> str:
        if not len(node.path):
            return f"{node.type} {node.target}"
        return f"{node.type} {node.target} {node.path}"

    @rule(IrLiteral)
    def literal(self, node: IrLiteral) -> str:
        return node.value.snbt()  # type: ignore

    @rule(IrStore)
    def store(self, node: IrStore) -> Generator[IrNode, str, str]:
        value = yield node.value
        type = node.type.value
        nbt_type = self.serialize_nbt_type(node.cast_type)

        match node.value:
            case IrScore():
                return f"execute store {type} score {value} run"
            case IrData():
                scale = number_to_string(node.scale)
                return f"execute store {type} {value} {nbt_type} {scale} run"
            case _:
                raise ValueError(f"Invalid store source '{node.value}'.")

    @rule(IrBinaryCondition, op="less_than")
    def less_than(self, node: IrBinaryCondition) -> Generator[IrNode, str, str]:
        left = yield node.left
        right = yield node.right
        test = "unless" if node.negated else "if"

        match node.left, node.right:
            case IrScore(), IrScore():
                return f"execute {test} score {left} < {right}"
            case IrScore(), IrLiteral() as r:
                value = cast(NumericNbtValue, r.value)
                right = str(value - 1)
                return f"execute {test} score {left} matches ..{right}"
            case IrLiteral() as l, IrScore():
                value = cast(NumericNbtValue, l.value)
                right = str(value + 1)
                return f"execute {test} score {right} matches {left}.."
            case l, r:
                raise InvalidOperand(node.op, l, r)

    @rule(IrBinaryCondition, op="greater_than")
    def greater_than(self, node: IrBinaryCondition) -> Generator[IrNode, str, str]:
        left = yield node.left
        right = yield node.right
        test = "unless" if node.negated else "if"

        match node.left, node.right:
            case IrScore(), IrScore():
                return f"execute {test} score {left} > {right}"
            case IrScore(), IrLiteral() as r:
                value = cast(NumericNbtValue, r.value)
                right = str(value + 1)
                return f"execute {test} score {left} matches {right}.."
            case IrLiteral() as l, IrScore():
                value = cast(NumericNbtValue, l.value)
                right = str(value - 1)
                return f"execute {test} score {right} matches ..{left}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

    @rule(IrBinaryCondition, op="greater_than_or_equal_to")
    def greater_than_or_qual_to(
        self, node: IrBinaryCondition
    ) -> Generator[IrNode, str, str]:
        left = yield node.left
        right = yield node.right
        test = "unless" if node.negated else "if"

        match node.left, node.right:
            case IrScore(), IrScore():
                return f"execute {test} score {left} >= {right}"
            case IrScore(), IrLiteral():
                return f"execute {test} score {left} matches {right}.."
            case IrLiteral(), IrScore():
                return f"execute {test} score {right} matches ..{left}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

    @rule(IrBinaryCondition, op="less_than_or_equal_to")
    def less_than_or_qual_to(
        self, node: IrBinaryCondition
    ) -> Generator[IrNode, str, str]:
        left = yield node.left
        right = yield node.right
        test = "unless" if node.negated else "if"

        match node.left, node.right:
            case IrScore(), IrScore():
                return f"execute {test} score {left} <= {right}"
            case IrScore(), IrLiteral():
                return f"execute {test} score {left} matches ..{right}"
            case IrLiteral(), IrScore():
                return f"execute {test} score {right} matches {left}.."
            case l, r:
                raise InvalidOperand(node.op, l, r)

    @rule(IrBinaryCondition, op="equal")
    def equal(self, node: IrBinaryCondition) -> Generator[IrNode, str, str]:
        left = yield node.left
        right = yield node.right
        test = "unless" if node.negated else "if"

        match node.left, node.right:
            case IrScore(), IrScore():
                return f"execute {test} score {left} = {right}"
            case IrScore(), IrLiteral():
                return f"execute {test} score {left} matches {right}"
            case IrLiteral(), IrScore():
                return f"execute {test} score {right} matches {left}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

    @rule(IrUnaryCondition, op="boolean")
    def boolean(self, node: IrUnaryCondition) -> Generator[IrNode, str, str]:
        target = yield node.target
        test = "unless" if node.negated else "if"

        match node.target:
            case IrBoolScore():
                return f"execute {test} score {target} matches 1"
            case IrScore() if node.negated:
                return f"execute unless score {target} matches ..-1 unless score {target} matches 1.."
            case IrScore() if not node.negated:
                return f"execute if score {target} matches -2147483648.. unless score {target} matches 0"
            case IrData():
                return f"execute {test} data {target}"
            case t:
                raise InvalidOperand(node.op, t)

    @rule(IrBranch)
    def branch(self, node: IrBranch) -> Generator[IrNode, str, None]:
        target = yield node.target

        match node.target:
            case IrSource() as source:
                cmd = yield IrUnaryCondition(op="boolean", target=source)
            case IrCondition():
                cmd = target

        children = self(node.children)
        self.add_result(cmd, node.store, children)

    @rule(IrBinary, op="add")
    def add(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        match node.left, node.right:
            case IrScore(), IrLiteral():
                cmd = f"scoreboard players add {left} {right}"
            case IrScore(), IrScore():
                cmd = f"scoreboard players operation {left} += {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrBinary, op="sub")
    def sub(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        match node.left, node.right:
            case IrScore(), IrLiteral():
                cmd = f"scoreboard players remove {left} {right}"
            case IrScore(), IrScore():
                cmd = f"scoreboard players operation {left} -= {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrBinary, op="mul")
    @rule(IrBinary, op="div")
    @rule(IrBinary, op="mod")
    @rule(IrBinary, op="min")
    @rule(IrBinary, op="max")
    def binary_score_only(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        operator = {
            "mul": "*=",
            "div": "/=",
            "mod": "%=",
            "min": "<",
            "max": ">",
        }
        match node.left, node.right:
            case IrScore(), IrScore():
                cmd = f"scoreboard players operation {left} {operator[node.op]} {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrBinary, op="append")
    def append(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        match node.left, node.right:
            case IrData(), IrLiteral():
                cmd = f"data modify {left} append value {right}"
            case IrData(), IrDataString() as s:
                start, end = s.normalized_range
                cmd = f"data modify {left} append string {right} {start}"
                if end is not None:
                    cmd += f" {end}"

            case IrData(), IrData():
                cmd = f"data modify {left} append from {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrBinary, op="prepend")
    def prepend(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        match node.left, node.right:
            case IrData(), IrLiteral():
                cmd = f"data modify {left} prepend value {right}"
            case IrData(), IrDataString() as s:
                start, end = s.normalized_range
                cmd = f"data modify {left} prepend string {right} {start}"
                if end is not None:
                    cmd += f" {end}"
            case IrData(), IrData():
                cmd = f"data modify {left} prepend from {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrInsert)
    def insert(self, node: IrInsert) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right
        index = node.index

        match node.left, node.right:
            case IrData(), IrLiteral():
                cmd = f"data modify {left} insert {index} value {right}"
            case IrData(), IrDataString() as s:
                start, end = s.normalized_range
                cmd = f"data modify {left} insert {index} string {right} {start}"
                if end is not None:
                    cmd += f" {end}"
            case IrData(), IrData():
                cmd = f"data modify {left} insert {index} from {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrBinary, op="merge")
    def merge(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        match node.left, node.right:
            case IrData() as l, IrLiteral() if not len(l.path):
                cmd = f"data merge {left} {right}"
            case IrData() as l, IrLiteral() if len(l.path):
                cmd = f"data modify {left} merge value {right}"
            case IrData() as l, IrDataString() as s if len(l.path):
                start, end = s.normalized_range
                cmd = f"data modify {left} merge string {right} {start}"
                if end is not None:
                    cmd += f" {end}"
            case IrData() as l, IrData() if len(l.path):
                cmd = f"data modify {left} merge from {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrUnary, op="remove")
    def remove(self, node: IrUnary) -> Generator[IrNode, str, None]:
        target = yield node.target

        match node.target:
            case IrData():
                cmd = f"data remove {target}"
            case t:
                raise InvalidOperand(node.op, t)

        self.add_result(cmd, node.store)

    @rule(IrUnary, op="reset")
    def reset_op(self, node: IrUnary) -> Generator[IrNode, str, None]:
        target = yield node.target

        match node.target:
            case IrScore():
                cmd = f"scoreboard players reset {target}"
            case t:
                raise InvalidOperand(node.op, t)

        self.add_result(cmd, node.store)

    @rule(IrUnary, op="enable")
    def enable(self, node: IrUnary) -> Generator[IrNode, str, None]:
        target = yield node.target

        match node.target:
            case IrScore():
                cmd = f"scoreboard players enable {target}"
            case t:
                raise InvalidOperand(node.op, t)

        self.add_result(cmd, node.store)

    def serialize_nbt_type(self, value: Any) -> NbtTypeString:
        value = unwrap_optional_type(value)

        if isinstance(value, str) and value in NBT_TYPE_STRING:
            return value

        if not isinstance(value, type):
            return self.default_nbt_type

        if issubclass(value, Byte):
            return "byte"
        if issubclass(value, Short):
            return "short"
        if issubclass(value, Long):
            return "long"
        if issubclass(value, Double):
            return "double"
        if issubclass(value, Float):
            return "float"
        if issubclass(value, Int):
            return "int"

        return self.default_nbt_type

    def serialize_cast(self, data: IrData) -> tuple[str, str]:
        cast_type = data.nbt_type
        scale = data.scale if data.scale is not None else 1

        if not isinstance(cast_type, str):
            cast_type = self.serialize_nbt_type(cast_type)

        if cast_type is None:
            cast_type = self.default_nbt_type

        return (cast_type, str(scale))

    @rule(IrSet)
    def set(self, node: IrBinary) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        match node.left, node.right:
            case IrScore(), IrLiteral():
                cmd = f"scoreboard players set {left} {right}"
            case IrScore(), IrScore():
                cmd = f"scoreboard players operation {left} = {right}"
            case IrData(), IrLiteral():
                cmd = f"data modify {left} set value {right}"
            case IrData(), IrDataString() as s:
                start, end = s.normalized_range
                cmd = f"data modify {left} set string {right} {start}"
                if end is not None:
                    cmd += f" {end}"
            case IrData(), IrData():
                cmd = f"data modify {left} set from {right}"
            case IrScore(), IrCondition():
                cmd = f"execute store success score {left} run {right}"
            case IrData(), IrCondition():
                cmd = f"execute store success {left} byte 1 run {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrCast, op="cast")
    def cast(self, node: IrCast) -> Generator[IrNode, str, None]:
        left = yield node.left
        right = yield node.right

        nbt_type = self.serialize_nbt_type(node.cast_type)
        scale = number_to_string(node.scale)

        match node.left, node.right:
            case IrData(), IrLiteral():
                cmd = f"data modify {left} set value {right}"
            case IrData() as l, IrData() as r:
                cmd = f"execute store result {left} {nbt_type} {scale} run data get {right} 1"
            case IrData() as l, IrScore():
                cmd = f"execute store result {left} {nbt_type} {scale} run scoreboard players get {right}"
            case IrData(), IrCondition():
                cmd = f"execute store success {left} {nbt_type} {scale} run {right}"
            case IrScore(), IrLiteral():
                cmd = f"scoreboard players set {left} {right}"
            case IrScore(), IrData() as r:
                cmd = f"execute store result score {left} run data get {right} {scale}"
            case IrScore(), IrCondition():
                cmd = f"execute store success score {left} run {right}"
            case l, r:
                raise InvalidOperand(node.op, l, r)

        self.add_result(cmd, node.store)

    @rule(IrUnary, op="get_length")
    def get_length(self, node: IrUnary) -> Generator[IrNode, str, None]:
        target = yield node.target

        match node.target:
            case IrData():
                cmd = f"data get {target}"
            case t:
                raise InvalidOperand(node.op, t)

        self.add_result(cmd, node.store)



from dataclasses import dataclass, replace
from typing import Any, Iterable, Tuple

from beet import Context
from beet.core.utils import required_field
from bolt import InterpolationParser, Runtime
from bolt.utils import internal
from mecha import (
    AstChildren,
    AstCommand,
    AstNode,
    CommandPrototype,
    CommandSpec,
    CommandTree,
    Mecha,
    MutatingReducer,
    rule,
)
from mecha.contrib.implicit_execute import ImplicitExecuteParser
from tokenstream import set_location

from bolt_expressions.node import Expression
from bolt_expressions.sources import DataSource, ScoreSource, Source

DEFAULT_PREFIX = "var"


def beet_default(ctx: Context):
    ctx.inject(commands)


def commands(ctx: Context):
    mc = ctx.inject(Mecha)
    runtime = ctx.inject(Runtime)

    mc.spec.parsers["command:argument:bolt_expressions:source"] = InterpolationParser(
        "source"
    )

    runtime.helpers["interpolate_source"] = SourceConverter(ctx=ctx)

    mc.transform.extend(SourceTransformer(mc=mc))

    prototypes = update_tree(mc.spec)

    update_implicit_execute(mc.spec, prototypes)


def update_tree(spec: CommandSpec):
    tree = spec.tree

    data_targets = (
        "minecraft:entity",
        "minecraft:block_pos",
        "minecraft:resource_location",
    )

    commands = spec.prototypes.keys()

    for command in commands:
        parsers = get_parsers(tree, command)
        scope = command.split(":")

        count = 0

        for i in range(0, len(parsers)):
            match parsers[: i + 1]:
                case [*_, "minecraft:score_holder", "minecraft:objective"]:
                    parent = tree.get(scope[: i - 1])
                    last_node = tree.get(scope[: i + 1])
                case [*_, None, target, "minecraft:nbt_path"] if target in data_targets:
                    parent = tree.get(scope[: i - 2])
                    last_node = tree.get(scope[: i + 1])
                case _:
                    continue

            prefix = DEFAULT_PREFIX
            name = f"sourceValue{count}"
            count += 1

            if prefix in parent.children:
                continue

            children = last_node.children

            end = CommandTree(
                type="argument",
                parser="bolt_expressions:source",
                properties={"prefix": prefix},
                redirect=last_node.redirect,
                executable=last_node.executable,
                subcommand=last_node.subcommand,
                children=None if children is None else dict(children),
            )
            parent.children[prefix] = CommandTree(type="literal", children={name: end})

    previous = set(spec.prototypes)

    spec.update()

    return set(spec.prototypes) - previous


def update_implicit_execute(spec: CommandSpec, prototypes: Iterable[CommandPrototype]):
    parser = spec.parsers["command"]
    while not isinstance(parser, ImplicitExecuteParser):
        parser = parser.parser

    keywords = {name.split(":", 2)[1] for name in parser.commands}

    commands = {
        name
        for name in prototypes
        if name.startswith("execute") and name.split(":", 2)[1] in keywords
    }
    shorthands = {
        name
        for name in prototypes
        if not name.startswith("execute") and name.split(":", 1)[0] not in keywords
    }

    parser.commands.update(commands)
    parser.shorthands.update(shorthands)


def get_parsers(tree: CommandTree, identifier: str):
    parsers = ()
    node = tree

    for name in identifier.split(":"):
        node = node.get(name)

        if node is None:
            break

        parsers += (node.parser,)

    return parsers


def get_child_argument_by_parser(node: CommandTree, parser: str):
    for name, child in node.children.items():
        if child.parser == parser:
            return name, child


def get_source_parsers(source: Source):
    if isinstance(source, ScoreSource):
        return ("score_holder", "objective")

    if isinstance(source, DataSource):
        if source._type == "entity":
            target = "entity"
        elif source._type == "storage":
            target = "resource_location"
        elif source._type == "block":
            target = "block_pos"
        else:
            target = None

        return (None, target, "nbt_path")


def get_source_values(source: Source):
    if isinstance(source, ScoreSource):
        return (source.scoreholder, source.objective)

    if isinstance(source, DataSource):
        return (source._type, source._target, str(source._path))


@dataclass(frozen=True)
class AstSourceNode(AstNode):
    value: Source = required_field()


@dataclass
class SourceConverter:
    ctx: Context | Expression

    @internal
    def __call__(self, obj: Any, node: AstNode):
        if isinstance(obj, AstNode):
            return set_location(obj, node.location, node.end_location)

        if isinstance(obj, Source):
            source = obj.evaluate()
        elif isinstance(obj, tuple) and len(obj) == 2:
            source = ScoreSource(*obj, ctx=self.ctx)
        elif isinstance(obj, tuple) and len(obj) == 3:
            source = DataSource(*obj, ctx=self.ctx)
        else:
            raise ValueError(
                f"Cannot interpolate source of type {type(obj)!r} '{obj}'."
            )

        return AstSourceNode(
            value=source, location=node.location, end_location=node.end_location
        )


@dataclass
class SourceTransformer(MutatingReducer):
    mc: Mecha = required_field()

    def get_argument_parent(
        self, command: AstCommand, arg_i: int
    ) -> Tuple[CommandTree, Tuple[str]]:
        spec = self.mc.spec

        prototype = spec.prototypes.get(command.identifier)
        prototype_arg = prototype.get_argument(arg_i)
        scope = prototype_arg.scope

        tree_arg = spec.tree.get(scope)

        properties = tree_arg.properties
        prefix = properties.get("prefix") if properties else None

        parent_scope = scope[:-1]
        offset = (prototype_arg.name,)

        if prefix and parent_scope[-1] == prefix:
            parent_scope = parent_scope[:-1]
            offset = (prefix, *offset)

        return spec.tree.get(parent_scope), offset

    @rule(AstCommand)
    def command(self, cmd_node: AstCommand):
        if not any(isinstance(n, AstSourceNode) for n in cmd_node.arguments):
            return cmd_node

        identifier = cmd_node.identifier
        arguments = []

        for i, cmd_arg in enumerate(cmd_node.arguments):
            if not isinstance(cmd_arg, AstSourceNode):
                arguments.append(cmd_arg)
                continue

            parent, parent_offset = self.get_argument_parent(cmd_node, i)

            source = cmd_arg.value

            parsers = get_source_parsers(source)
            values = get_source_values(source)

            arg_scope = []
            arg_parent = parent

            for parser, value in zip(parsers, values):
                if parser is None:
                    arg_scope.append(value)
                    arg_parent = arg_parent.get(value)
                    continue

                arg_name, arg_node = get_child_argument_by_parser(
                    arg_parent, parser if ":" in parser else f"minecraft:{parser}"
                )

                if not value:
                    continue

                arg_ast = self.mc.parse(
                    value, using=parser, provide={"properties": arg_node.properties}
                )

                arguments.append(arg_ast)
                arg_scope.append(arg_name)

                arg_parent = arg_node

            identifier = identifier.replace(
                ":".join(parent_offset), ":".join(arg_scope), 1
            )

        arguments = AstChildren(arguments)

        return replace(cmd_node, identifier=identifier, arguments=arguments)



from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from typing import Any, ClassVar, Iterable

from .literals import convert_node
from .node import ExpressionNode, ResultType, UnrollHelper
from .optimizer import (
    IrBinary,
    IrBinaryCondition,
    IrCast,
    IrChildren,
    IrCondition,
    IrData,
    IrGetLength,
    IrInsert,
    IrLiteral,
    IrNode,
    IrOperation,
    IrScore,
    IrSet,
    IrSource,
    IrStore,
    IrUnary,
    IrUnaryCondition,
    StoreType,
)
from .typing import NbtType

__all__ = [
    "ResultType",
    "Operation",
    "UnaryOperation",
    "BinaryOperation",
    "Remove",
    "Merge",
    "InPlaceMerge",
    "Insert",
    "Append",
    "Prepend",
    "Set",
    "Enable",
    "Reset",
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Modulus",
    "Min",
    "Max",
]


@dataclass(unsafe_hash=False, order=False, kw_only=True)
class Operation(ExpressionNode, ABC):
    in_place: ClassVar[bool] = False
    evaluates_target: ClassVar[bool] = True
    commutative: ClassVar[bool] = False
    op: ClassVar[str] = ""
    result: ClassVar[ResultType] = ResultType.score
    stores_result: ClassVar[bool] = False

    @property
    @abstractmethod
    def in_place_target(self) -> ExpressionNode | None: ...


@dataclass(unsafe_hash=False, order=False)
class UnaryOperation(Operation):
    target: ExpressionNode

    @property
    def in_place_target(self) -> ExpressionNode | None:
        if not self.in_place:
            return None

        return self.target

    def create_operation(self, target: IrSource) -> IrUnary:
        return IrUnary(op=self.op, target=target)

    def unroll(self, helper: UnrollHelper) -> tuple[Iterable[IrOperation], IrSource]:
        target = convert_node(self.target, self.ctx)

        with helper.provide(ignore_lazy=not self.evaluates_target):
            target_nodes, target_value = target.unroll(helper)

        if not isinstance(target_value, IrSource):
            raise ValueError("Operand must be a source node.")

        operations: list[IrOperation] = [*target_nodes]

        if self.in_place:
            temp_var = target_value
        else:
            temp_var = helper.create_temporary(self.result)

        if self.stores_result:
            operation = self.create_operation(target_value)

            store = IrChildren(
                (*operation.store, IrStore(type=StoreType.result, value=temp_var))
            )
            operation = replace(operation, store=store)
        else:
            operations.append(IrSet(left=temp_var, right=target_value))
            operation = self.create_operation(temp_var)

        operations.append(operation)

        return operations, temp_var


def balance_priority(node: IrNode, helper: UnrollHelper) -> int:
    if isinstance(node, IrSource) and node.to_tuple() in helper.temporaries:
        return 3

    if isinstance(node, IrData):
        return 2

    if isinstance(node, IrScore):
        return 1

    return 0


@dataclass(unsafe_hash=False, order=False)
class BinaryOperation(Operation):
    former: ExpressionNode
    latter: Any

    @property
    def in_place_target(self) -> ExpressionNode | None:
        if not self.in_place:
            return None

        return self.former

    def create_operation(self, left: IrSource, right: IrSource | IrLiteral) -> IrBinary:
        return IrBinary(op=self.op, left=left, right=right)

    def unroll(self, helper: UnrollHelper) -> tuple[Iterable[IrOperation], IrSource]:
        former = convert_node(self.former, self.ctx)
        latter = convert_node(self.latter, self.ctx)

        with helper.provide(ignore_lazy=not self.evaluates_target):
            former_nodes, former_value = former.unroll(helper)
        latter_nodes, latter_value = latter.unroll(helper)

        if self.commutative:
            former_priority = balance_priority(former_value, helper)
            latter_priority = balance_priority(latter_value, helper)

            if former_priority < latter_priority:
                node = replace(self, former=latter, latter=former)
                return node.unroll(helper)

        operations: list[IrOperation] = [*former_nodes, *latter_nodes]

        if self.in_place and isinstance(former_value, IrSource):
            temp_var = former_value
        else:
            temp_var = helper.create_temporary(self.result)
            operations.append(IrSet(left=temp_var, right=former_value))

        operation = self.create_operation(temp_var, latter_value)

        if self.stores_result:
            store = IrChildren(
                (*operation.store, IrStore(type=StoreType.result, value=temp_var))
            )
            operation = replace(operation, store=store)

        operations.append(operation)

        return operations, temp_var


class Remove(UnaryOperation):
    op: ClassVar[str] = "remove"
    in_place: ClassVar[bool] = True


class Merge(BinaryOperation):
    op: ClassVar[str] = "merge"
    result: ClassVar[ResultType] = ResultType.data


class InPlaceMerge(BinaryOperation):
    op: ClassVar[str] = "merge"
    in_place: ClassVar[bool] = True


@dataclass(eq=False, order=False)
class Insert(BinaryOperation):
    op: ClassVar[str] = "insert"
    in_place: ClassVar[bool] = True

    index: int = 0

    def create_operation(
        self, left: IrSource, right: IrSource | IrLiteral | IrCondition
    ) -> IrInsert:
        return IrInsert(left=left, right=right, index=self.index)


class Append(BinaryOperation):
    op: ClassVar[str] = "append"
    in_place: ClassVar[bool] = True


class Prepend(BinaryOperation):
    op: ClassVar[str] = "prepend"
    in_place: ClassVar[bool] = True


class Set(BinaryOperation):
    op: ClassVar[str] = "set"
    in_place: ClassVar[bool] = True
    evaluates_target: ClassVar[bool] = False

    def create_operation(
        self, left: IrSource, right: IrSource | IrLiteral | IrCondition
    ) -> IrSet:
        return IrSet(left=left, right=right)


@dataclass(eq=False, order=False)
class Cast(BinaryOperation):
    op: ClassVar[str] = "cast"
    in_place: ClassVar[bool] = True
    evaluates_target: ClassVar[bool] = False

    cast_type: NbtType = Any

    def create_operation(
        self, left: IrSource, right: IrSource | IrLiteral | IrCondition
    ) -> IrCast:
        return IrCast(left=left, right=right, cast_type=self.cast_type)


class Enable(UnaryOperation):
    op: ClassVar[str] = "enable"
    in_place: ClassVar[bool] = True


class Reset(UnaryOperation):
    op: ClassVar[str] = "reset"
    in_place: ClassVar[bool] = True
    evaluates_target: ClassVar[bool] = False


class Add(BinaryOperation):
    op: ClassVar[str] = "add"
    commutative: ClassVar[bool] = True


class Subtract(BinaryOperation):
    op: ClassVar[str] = "sub"


class Multiply(BinaryOperation):
    op: ClassVar[str] = "mul"
    commutative: ClassVar[bool] = True


class Divide(BinaryOperation):
    op: ClassVar[str] = "div"


class Modulus(BinaryOperation):
    op: ClassVar[str] = "mod"


class Min(BinaryOperation):
    op: ClassVar[str] = "min"
    commutative: ClassVar[bool] = True


class Max(BinaryOperation):
    op: ClassVar[str] = "max"
    commutative: ClassVar[bool] = True


class GetLength(UnaryOperation):
    op: ClassVar[str] = "get_length"
    stores_result: ClassVar[bool] = True

    def create_operation(self, target: IrSource) -> IrUnary:
        return IrGetLength(target=target)


class UnaryCondition(UnaryOperation):
    negated: ClassVar[bool] = False

    def unroll(self, helper: UnrollHelper) -> tuple[Iterable[IrOperation], IrSource]:
        target_nodes, target_var = convert_node(self.target, self.ctx).unroll(helper)

        if not isinstance(target_var, IrSource):
            raise ValueError("Operand must be a source node.")

        condition = IrUnaryCondition(
            op=self.op, target=target_var, negated=self.negated
        )
        temp_var = helper.create_temporary(ResultType.score)
        op = IrSet(left=temp_var, right=condition)

        return (*target_nodes, op), temp_var


@dataclass
class BinaryCondition(BinaryOperation):
    negated: ClassVar[bool] = False

    def unroll(self, helper: UnrollHelper) -> tuple[Iterable[IrOperation], IrSource]:
        former_nodes, former_var = convert_node(self.former, self.ctx).unroll(helper)
        latter_nodes, latter_var = convert_node(self.latter, self.ctx).unroll(helper)

        temp_var = helper.create_temporary(ResultType.score)
        condition = IrBinaryCondition(
            op=self.op, left=former_var, right=latter_var, negated=self.negated
        )
        op = IrSet(left=temp_var, right=condition)

        return (*former_nodes, *latter_nodes, op), temp_var


class Boolean(UnaryCondition):
    op: ClassVar[str] = "boolean"


class Not(UnaryCondition):
    op: ClassVar[str] = "boolean"
    negated: ClassVar[bool] = True


class LessThan(BinaryCondition):
    op: ClassVar[str] = "less_than"


class LessThanOrEqualTo(BinaryCondition):
    op: ClassVar[str] = "less_than_or_equal_to"


class GreaterThan(BinaryCondition):
    op: ClassVar[str] = "greater_than"


class GreaterThanOrEqualTo(BinaryCondition):
    op: ClassVar[str] = "greater_than_or_equal_to"


class Equal(BinaryCondition):
    op: ClassVar[str] = "equal"


class NotEqual(BinaryCondition):
    op: ClassVar[str] = "equal"
    negated: ClassVar[bool] = True



__all__ = [
    "TypeCheckError",
    "TypeCheckDiagnostic",
    "get_exception_chain",
]


class TypeCheckError(Exception):
    """Type incompatibility during type checking"""


class TypeCheckDiagnostic(Exception):
    """Diagnostic error raised by the type checker."""


def get_exception_chain(exc: BaseException) -> tuple[BaseException, ...]:
    exceptions: list[BaseException] = []

    current_exc = exc

    while current_exc:
        exceptions.append(current_exc)
        current_exc = current_exc.__cause__

    return tuple(exceptions)



from contextlib import suppress
from dataclasses import dataclass, replace
from types import NoneType
from typing import Any, Iterable

from beet import Context
from mecha import Visitor, rule
from nbtlib import (  # type: ignore
    Array,
    Compound,
    List,
    ListIndex,
    NamedKey,
    OutOfRange,
    String,
)

from .optimizer import IrBinary, IrCast, IrData, IrLiteral, IrOperation
from .typing import (
    NbtType,
    NbtValue,
    NumericNbtValue,
    access_type,
    convert_tag,
    is_array_type,
    is_compound_type,
    is_list_type,
    is_numeric_type,
    is_string_type,
    is_union,
    unwrap_optional_type,
)

__all__ = [
    "TypeCaster",
]


def cast_dict(
    nbt_type: NbtType, value: dict[Any, Any], ctx: Context | None = None
) -> Compound | None:
    if not is_compound_type(nbt_type):
        return None

    result: dict[Any, Any] = {}

    for key, val in value.items():
        key_type = access_type(nbt_type, NamedKey(key), ctx)

        if not key_type:
            result[key] = val
            continue

        casted_val = cast_value(key_type, val, ctx)

        result[key] = casted_val if casted_val is not None else val

    return Compound(result)


def cast_list(
    nbt_type: NbtType, value: list[Any] | Array, ctx: Context | None = None
) -> List | Array | None:
    if is_list_type(nbt_type):
        cast_type = List
    elif is_array_type(nbt_type):
        cast_type = nbt_type
    else:
        return None

    result: list[Any] = []

    for i, element in enumerate(value):
        el_type = access_type(nbt_type, ListIndex(i))

        if el_type is None:
            return None

        element = cast_value(el_type, element, ctx)

        if element is None:
            return None

        result.append(element)

    return cast_type(result)


def cast_numeric(nbt_type: NbtType, value: int | float) -> NumericNbtValue | None:
    if not is_numeric_type(nbt_type):
        return None

    with suppress(OutOfRange):
        return nbt_type(value)


def cast_string(nbt_type: NbtType, value: str) -> String | None:
    if is_string_type(nbt_type):
        return String(value)

    return None


def cast_value(
    nbt_type: NbtType, value: NbtValue | Any, ctx: Context | None = None
) -> NbtValue | None:
    nbt_type = unwrap_optional_type(nbt_type)

    if nbt_type in (Any, None, NoneType) or is_union(nbt_type):
        return convert_tag(value)

    if isinstance(value, dict):
        return cast_dict(nbt_type, value, ctx)

    if isinstance(value, list):
        return cast_list(nbt_type, value, ctx)

    if isinstance(value, (int, float)):
        return cast_numeric(nbt_type, value)

    if isinstance(value, str):
        return cast_string(nbt_type, value)

    return convert_tag(value)


@dataclass(eq=False, kw_only=True)
class TypeCaster(Visitor):
    ctx: Context | None

    def __call__(self, nodes: Iterable[IrOperation]) -> Iterable[IrOperation]:  # type: ignore
        for node in nodes:
            yield self.invoke(node)

    @rule(IrOperation)
    def fallback(self, node: IrOperation) -> IrOperation:
        return node

    @rule(IrCast)
    def cast(self, node: IrCast) -> IrBinary:
        if not isinstance(node.left, IrData):
            return node
        if not isinstance(node.right, IrLiteral):
            return node

        casted_value = cast_value(node.cast_type, node.right.value, self.ctx)

        if casted_value is None:
            return node

        return replace(node, right=IrLiteral(value=casted_value))

    @rule(IrBinary, op="merge")
    def set(self, node: IrBinary) -> IrBinary:
        if not isinstance(node.left, IrData):
            return node
        if not isinstance(node.right, IrLiteral):
            return node

        casted_value = cast_value(node.left.nbt_type, node.right.value, self.ctx)

        if casted_value is None:
            return node

        return replace(node, right=IrLiteral(value=casted_value))

    @rule(IrBinary, op="insert")
    @rule(IrBinary, op="append")
    @rule(IrBinary, op="prepend")
    def insert(self, node: IrBinary) -> IrBinary:
        if not isinstance(node.left, IrData):
            return node
        if not isinstance(node.right, IrLiteral):
            return node

        nbt_type = access_type(node.left.nbt_type, ListIndex(None), self.ctx)

        if nbt_type is None:
            return node

        casted_value = cast_value(nbt_type, node.right.value, self.ctx)

        if casted_value is None:
            return node

        return replace(node, right=IrLiteral(value=casted_value))



from dataclasses import dataclass, field
from functools import partial
from typing import Any

from beet import Context
from bolt import Runtime
from bolt.utils import internal
from mecha import Mecha

import bolt_expressions as bolt_expressions_module

from .api import Data, Scoreboard
from .ast import (
    ConstantScoreChecker,
    ObjectiveChecker,
    RunExecuteTransformer,
    SourceNBTConverter,
)
from .expose import wrapped_len, wrapped_max, wrapped_min
from .node import Expression

__all__ = [
    "beet_default",
    "bolt_expressions",
]


def bolt_expressions(ctx: Context):
    ctx.require("bolt_control_flow")

    expr = ctx.inject(Expression)
    scoreboard = ctx.inject(Scoreboard)
    data = ctx.inject(Data)

    mc = ctx.inject(Mecha)
    mc.transform.extend(RunExecuteTransformer())
    mc.check.extend(
        ConstantScoreChecker(
            objective=expr.opts.const_objective, callback=scoreboard.add_constant
        ),
        ObjectiveChecker(
            whitelist=scoreboard.objectives,
            callback=scoreboard.add_objective,
        ),
    )

    runtime = ctx.inject(Runtime)

    api = {
        "Expression": expr,
        "Scoreboard": scoreboard,
        "Data": data,
    }
    handler = module_attribute_handler(
        ctx, runtime.helpers["get_attribute_handler"], api
    )
    nbt_converter = SourceNBTConverter(runtime.helpers["interpolate_nbt"])
    runtime.helpers["get_attribute_handler"] = handler
    runtime.helpers["interpolate_nbt"] = nbt_converter

    runtime.expose("min", partial(wrapped_min, runtime.globals.get("min", min)))
    runtime.expose("max", partial(wrapped_max, runtime.globals.get("max", max)))
    runtime.expose("len", partial(wrapped_len, runtime.globals.get("len", len)))

    if not expr.opts.disable_commands:
        ctx.require("bolt_expressions.contrib.commands")

    yield

    expr.generate_init()


beet_default = bolt_expressions


def module_attribute_handler(
    ctx: Context, previus_handler: Any, attributes: dict[str, Any]
):
    def handler(obj: Any):
        if obj is bolt_expressions_module:
            return ExtendedAttributeHandler(obj, previus_handler(obj), ctx, attributes)

        return previus_handler(obj)

    return handler


@dataclass
class ExtendedAttributeHandler:
    obj: Any
    handler: Any
    ctx: Context

    attributes: dict[str, Any] = field(default_factory=dict)

    @internal
    def __getitem__(self, attr: str) -> Any:
        if attr in self.attributes:
            return self.attributes[attr]

        return self.handler[attr]

    @internal
    def __setitem__(self, attr: str, value: Any):
        self.handler[attr] = value

    @internal
    def __delitem__(self, attr: str):
        del self.handler[attr]



from typing import Any, Iterable, TypeVar, Union

from .node import ExpressionNode
from .operations import Max, Min
from .sources import Source, binary_operator, length

T = TypeVar("T")

binary_min = binary_operator(Min)
binary_max = binary_operator(Max)


def wrapped_min(f: Any, *args: T, **kwargs: Any) -> Union[T, Any]:
    values = args

    if len(args) == 1:
        if isinstance(args[0], ExpressionNode):
            return args[0]
        if not isinstance(args[0], Iterable):
            return args[0]

        values = tuple(args[0])

    for i, node in enumerate(values):
        if not isinstance(node, ExpressionNode):
            continue

        remaining = values[:i] + values[i + 1 :]
        return binary_min(node, wrapped_min(f, *remaining, **kwargs))

    return f(*args, **kwargs)


def wrapped_max(f: Any, *args: T, **kwargs: Any) -> Union[T, Any]:
    values = args

    if len(args) == 1:
        if isinstance(args[0], ExpressionNode):
            return args[0]
        if not isinstance(args[0], Iterable):
            return args[0]

        values = tuple(args[0])

    for i, node in enumerate(values):
        if not isinstance(node, ExpressionNode):
            continue

        remaining = values[:i] + values[i + 1 :]
        return binary_max(node, wrapped_max(f, *remaining, **kwargs))

    return f(*args, **kwargs)


def wrapped_len(f: Any, obj: Any, /) -> Any:
    if not isinstance(obj, Source):
        return f(obj)

    return length(obj)



from dataclasses import dataclass, field
from typing import Any, overload

from beet import Context
from bolt.utils import internal

from .node import Expression
from .sources import (
    DataSource,
    ScoreSource,
    resolve,
)
from .typing import NbtType, literal_types

__all__ = [
    "Scoreboard",
    "Objective",
    "Data",
]


class Scoreboard:
    """API for manipulating scoreboards.

    To use, inject the current `Context` and construct an `Score` instance.
    ```
        Objective = ctx.inject(Scoreboard)
        my_obj = Objective["my_obj"]
    ```
    Now you can perform the API manipulation via the operators:
    ```
        my_obj["@s"] += 10
        my_obj["temp"] = my_obj["@s"] * 10
        player = my_obj["@s"]
        player += 10 * my_obj["temp"]
    ```
    """

    expr: Expression

    constants: set[int]
    objectives: set[str]
    added_objectives: set[str]

    def __init__(self, ctx: Context | Expression):
        if isinstance(ctx, Context):
            self.expr = ctx.inject(Expression)
        else:
            self.expr = ctx

        opts = self.expr.opts

        self.constants = set()
        self.objectives = {opts.const_objective, opts.temp_objective}
        self.added_objectives = set()

    def add_objective(self, name: str, criteria: str = "dummy"):
        if name not in self.added_objectives:
            self.added_objectives.add(name)

            self.expr.init_commands.insert(
                0, f"scoreboard objectives add {name} {criteria}"
            )

    def add_constant(self, value: int):
        holder, obj = self.expr.const_score(value)

        if value not in self.constants:
            self.constants.add(value)
            self.expr.init_commands.append(
                f"scoreboard players set {holder} {obj} {value}"
            )

    def objective(
        self, name: str, criteria: str | None = None, prefixed: bool = True
    ) -> "Objective":
        """
        Get an Objective instance and add objective to init function if provided criteria.
        """

        if prefixed:
            name = self.expr.opts.objective_prefix + name

        if criteria:
            self.add_objective(name, criteria)

        return Objective(name, ctx=self.expr)

    __call__ = objective

    def score(self, scoreholder: str, objective: str) -> ScoreSource:
        return ScoreSource(scoreholder, objective, ctx=self.expr)


@dataclass
class Objective:
    name: str
    ctx: Context | Expression = field(kw_only=True)

    @overload
    def __getitem__(self, scoreholder: str) -> ScoreSource: ...

    @overload
    def __getitem__(self, scoreholder: tuple[str, ...]) -> tuple[ScoreSource, ...]: ...

    def __getitem__(
        self, scoreholder: str | tuple[str, ...]
    ) -> ScoreSource | tuple[ScoreSource, ...]:
        if isinstance(scoreholder, str):
            return ScoreSource(scoreholder, self.name, ctx=self.ctx)

        return tuple(
            ScoreSource(holder, self.name, ctx=self.ctx) for holder in scoreholder
        )

    @internal
    def __setitem__(self, scoreholder: str | tuple[str, ...], value: Any):
        target = self[scoreholder]

        scores = target if isinstance(target, tuple) else (target,)
        for score in scores:
            score.__rebind__(value)

    def __str__(self):
        return self.name


class Data:
    expr: Expression

    def __init__(self, ctx: Context | Expression):
        if isinstance(ctx, Context):
            self.expr = ctx.inject(Expression)
        else:
            self.expr = ctx

    def __call__(self, target: str):
        """Guess target type and return a data source."""
        ...

    def storage(self, resource_location: str):
        return DataSource("storage", resource_location, ctx=self.expr)

    def entity(self, entity: str):
        return DataSource("entity", entity, ctx=self.expr)

    def block(self, position: str):
        return DataSource("block", position, ctx=self.expr)

    def dummy(self, type: NbtType | str = Any):
        "Create a dummy data source in a storage."

        if isinstance(type, str):
            type = literal_types[type]

        target_type, target, path = self.expr.temp_data()
        return DataSource(target_type, target, path, ctx=self.expr)[type]

    def cast(self, value: Any, nbt_type: Any = Any):
        if isinstance(nbt_type, str):
            nbt_type = literal_types[nbt_type]

        result = self.dummy(nbt_type)
        resolve(self.expr, value, result, nbt_type, lazy=True)

        return result



from dataclasses import dataclass, field
from typing import Any, Iterable, cast

from beet import Context
from beet.core.utils import required_field

from .node import Expression, ExpressionNode, UnrollHelper
from .optimizer import (
    CompositeNbtValue,
    IrCompositeLiteral,
    IrLiteral,
    IrOperation,
    IrSource,
)
from .typing import NbtValue, convert_tag
from .utils import type_name

__all__ = [
    "Literal",
    "convert_node",
]


@dataclass(unsafe_hash=True, order=False)
class Literal(ExpressionNode):
    value: Any = required_field()
    nbt: NbtValue | None = field(init=False)

    def __post_init__(self):
        self.nbt = convert_tag(self.value)

    def unroll(self, helper: UnrollHelper) -> tuple[Iterable[IrOperation], Any]:
        operations: list[IrOperation] = []

        def nested_unroll(obj: Any) -> CompositeNbtValue | IrSource:
            if isinstance(obj, dict):
                obj = cast(dict[str, Any], obj)
                return {key: nested_unroll(value) for key, value in obj.items()}

            if isinstance(obj, list):
                obj = cast(list[Any], obj)
                return [nested_unroll(value) for value in obj]

            if isinstance(obj, ExpressionNode):
                ops, value = obj.unroll(helper)
                operations.extend(ops)

                return value.value if isinstance(value, IrLiteral) else value

            value = convert_tag(obj)

            if value is None:
                raise ValueError(
                    f'Invalid literal of type {type_name(obj)} "{self.value}"'
                )

            return value

        if self.nbt is None:
            value = nested_unroll(self.value)

            if isinstance(value, IrSource):
                return tuple(operations), value

            return tuple(operations), IrCompositeLiteral(value=value)

        return (), IrLiteral(value=self.nbt)


def convert_node(value: Any, ctx: Context | Expression) -> ExpressionNode:
    if isinstance(value, ExpressionNode):
        return value

    return Literal(value=value, ctx=ctx)



import sys
from contextlib import contextmanager
from dataclasses import replace
from types import NoneType
from typing import Any, Dict, is_typeddict  # type: ignore

from beet import Context
from bolt import Runtime
from mecha import AstChildren, AstCommand, AstRoot
from nbtlib import Base  # type: ignore

__all__ = [
    "type_name",
    "identifier_generator",
    "get_globals",
    "assert_exception",
]


def type_name(obj: Any) -> str:
    return type(obj).__name__


def format_name(t: Any) -> str:
    if not isinstance(t, type):
        return repr(t)

    if is_typeddict(t):
        return f"{t.__module__}.{t.__name__}"

    if issubclass(t, (bool, int, float, str, list, dict, NoneType, Base)):
        return t.__name__

    return f"{t.__module__}.{t.__name__}"


def identifier_generator(ctx: Context | None = None):
    if ctx:
        runtime = ctx.inject(Runtime)
        incr: Dict[str, int] = {}

        while True:
            path = runtime.modules.current_path
            incr[path] = incr.setdefault(path, -1) + 1

            yield ctx.generate.format(f"{{hash}}_{incr[path]}", path)
    else:
        counter = 0
        while True:
            yield str(f"i{counter}")


def get_globals(obj: Any, ctx: Context | Runtime | None = None) -> dict[str, Any]:
    if isinstance(ctx, Context):
        runtime = ctx.inject(Runtime)
    else:
        runtime = ctx

    if runtime is not None:
        module = runtime.modules.get(obj.__module__)
        if module:
            return module.namespace

    return getattr(sys.modules.get(obj.__module__, None), "__dict__", {})


@contextmanager
def assert_exception(exc: type[Exception]):
    try:
        yield
    except exc:
        ...
    else:
        raise AssertionError(f"Expected {exc.__name__} to be raised.")


def insert_nested_commands(execute: AstCommand, root: AstRoot) -> AstCommand:
    """Inserts nested commands to the end of an execute command."""

    subcommand = None

    if len(execute.arguments):
        subcommand = execute.arguments[-1]

    if isinstance(subcommand, AstRoot):
        return replace(execute, arguments=AstChildren((*execute.arguments[:-1], root)))

    if not isinstance(subcommand, AstCommand):
        commands = AstCommand(
            identifier="execute:commands", arguments=AstChildren((root,))
        )
        arguments = AstChildren((*execute.arguments, commands))
        identifier = execute.identifier + ":subcommand"
        return replace(execute, identifier=identifier, arguments=arguments)

    inserted = insert_nested_commands(subcommand, root)
    arguments = AstChildren((*execute.arguments[:-1], inserted))
    return replace(execute, arguments=arguments)



import re
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, Iterable, cast

from beet.core.utils import required_field
from mecha import (
    AstCommand,
    AstNode,
    AstObjective,
    AstPlayerName,
    MutatingReducer,
    Reducer,
    rule,
)
import nbtlib

from .sources import Source

__all__ = [
    "RunExecuteTransformer",
    "ConstantScoreChecker",
    "ObjectiveChecker",
    "SourceNBTConverter",
]


@dataclass
class RunExecuteTransformer(MutatingReducer):
    @rule(AstCommand, identifier="execute:run:subcommand")
    def strip_run_execute(self, node: AstCommand):
        if isinstance(command := node.arguments[0], AstCommand):
            if command.identifier == "execute:subcommand":
                return command.arguments[0]
        return node


@dataclass
class ConstantScoreChecker(Reducer):
    objective: str = required_field()
    callback: Callable[[int], None] = required_field()
    pattern: re.Pattern[str] = re.compile(r"^[$#]([-+]?\d+)\b")

    @cached_property
    def objective_node(self):
        return AstObjective(value=self.objective)

    @rule(AstCommand)
    def command(self, node: AstCommand):
        if self.objective_node not in node.arguments:
            return

        i = node.arguments.index(self.objective_node)
        name = node.arguments[i - 1]

        if not isinstance(name, AstPlayerName):
            return

        match = self.pattern.match(name.value)

        if not match:
            return

        value = int(match.group(1))
        self.callback(value)


@dataclass
class ObjectiveChecker(Reducer):
    whitelist: Iterable[str] = required_field()
    callback: Callable[[str], None] = required_field()

    @rule(AstObjective)
    def objective(self, node: AstObjective):
        value = node.value

        if value not in self.whitelist:
            return

        self.callback(value)


@dataclass
class SourceNBTConverter:
    converter: Callable[[Any, AstNode], AstNode]

    def convert(self, obj: Any) -> Any:
        if isinstance(obj, Source):
            return self.convert(obj.component())

        if isinstance(obj, (list, tuple)):
            list_value = cast(list[Any], obj)
            return nbtlib.List([self.convert(value) for value in list_value])

        if isinstance(obj, dict):
            dict_value = cast(dict[str, Any], obj)
            return nbtlib.Compound({key: self.convert(value) for key, value in dict_value.items()})

        return obj

    def __call__(self, obj: Any, node: AstNode):
        obj = self.convert(obj)

        return self.converter(obj, node)
